# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shutil
import tempfile
import unittest
from os import path

import pyvelox.pyvelox as pv


class TestVeloxVectorSaver(unittest.TestCase):
    def setUp(self):
        # create a temporary directory
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        # remove the temporary directory
        shutil.rmtree(self.test_dir)

    def make_flat_vector(self):
        return pv.from_list([1, 2, 3])

    def make_const_vector(self):
        return pv.constant_vector(1000, 10)

    def make_dict_vector(self):
        base_indices = [0, 0, 1, 0, 2]
        return pv.dictionary_vector(pv.from_list([1, 2, 3]), base_indices)

    def test_serde_vector(self):
        data = {
            "flat_vector": self.make_flat_vector(),
            "const_vector": self.make_const_vector(),
            "dict_vector": self.make_dict_vector(),
        }

        paths = {
            "flat_vector": path.join(self.test_dir, "flat.pyvelox"),
            "const_vector": path.join(self.test_dir, "const.pyvelox"),
            "dict_vector": path.join(self.test_dir, "dict.pyvelox"),
        }

        for vec_key, fpath_key in zip(data, paths):
            vec = data[vec_key]
            fpath = paths[fpath_key]
            pv.save_vector(vector=vec, file_path=fpath)
            loaded_vec = pv.load_vector(file_path=fpath)
            self.assertEqual(len(vec), len(loaded_vec))
            self.assertEqual(vec.dtype(), loaded_vec.dtype())
            for i in range(len(vec)):
                self.assertEqual(vec[i], loaded_vec[i])



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

import unittest
import pyarrow

from velox.py.arrow import to_velox, to_arrow
from velox.py.vector import Vector


class TestPyVeloxArrow(unittest.TestCase):
    def test_vector_simple(self):
        array = pyarrow.array([1, 2, 3, 4, 5, 6])
        vector = to_velox(array)

        self.assertTrue(isinstance(vector, Vector))
        self.assertEqual(len(vector), 6)

        # TODO: For now we only return the values as strings for printing.
        for i in range(len(array)):
            self.assertEqual(vector[i], str(array[i]))

    def test_roundtrip(self):
        array = pyarrow.array([2, 2, 3, 4, 4, 0])
        array2 = to_arrow(to_velox(array))

        self.assertTrue(isinstance(array2, pyarrow.Array))
        self.assertEqual(array, array2)

    def test_empty(self):
        # TODO: Velox's arrow bridge does not allow missing buffers (even if
        # there are no rows):
        #   https://github.com/facebookincubator/velox/issues/12082

        # vector = to_velox(pyarrow.array([]))
        # self.assertEqual(vector.size(), 0)
        pass



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
import os
import sys

import util

CIRCLECI_V2 = "https://circleci.com/api/v2"
CIRCLECI_V1 = "https://circleci.com/api/v1.1"


def get_circleci_token():
    token = os.getenv("CIRCLECI_API_TOKEN")

    if token is None:
        print(
            """
            CIRCLECI_API_TOKEN is not set in the environment.  An API key is
            required to access CircleCi.  You can create an API token here :

                https://circleci.com/account/api

            Then set this value in your .profile or .bashrc.

                export CIRCLECI_API_TOKEN="Your Token Value Here"
""",
            file=sys.stderr,
        )

        sys.exit(1)

    return token


def circleci_url(token, url, compressed=False):
    return util.run(
        f"curl -s --header 'Circle-Token: {token}' \
       --header 'Accept: application/json'    \
       --header 'Content-Type: application/json' \
       '{url}'",
        compressed=compressed,
    )[1]


def get_pr_workflows(token, repo, pr):
    url = f"{CIRCLECI_V2}/insights/gh/{repo}/workflows/dist-compile?branch=pull/{pr}"

    return circleci_url(token, url).json()


def get_workflow_jobs(token, id):
    url = f"{CIRCLECI_V2}/workflow/{id}/job"

    return circleci_url(token, url).json()


def get_job_steps(token, repo, job):
    url = f"{CIRCLECI_V1}/project/gh/{repo}/{job}"

    return circleci_url(token, url).json()


def get_action_output(token, url):
    return circleci_url(token, url, compressed=True).json()


def output(args):
    github_upstream = util.run("git remote get-url --push upstream")[1].extract(
        r".*:(.*)[.]git"
    )
    github_user = util.run("git remote get-url --push origin")[1].extract(r".*:(.*)/")
    git_branch = util.run("git symbolic-ref --short HEAD")[1]

    if args.pr:
        pull_request = args.pr
    else:
        stdout = util.run(f"hub pr list -h {github_user}:{git_branch}")[1]
        if stdout == "":
            print(f"Error: no pull request for {github_user}:{git_branch}")
            sys.exit(1)

        pull_request = stdout.extract(r".*#([0-9]+) .*")

    if args.token is None:
        token = get_circleci_token()
    else:
        token = args.token

    pr_workflows = get_pr_workflows(token, github_upstream, pull_request)
    latest_workflow = sorted(
        pr_workflows["items"], key=lambda x: x.created_at, reverse=True
    )[0]
    latest_workflow_id = latest_workflow.id

    workflow_jobs = get_workflow_jobs(token, latest_workflow_id)

    for job in workflow_jobs["items"]:
        if job.status != args.job_status or args.job_name not in job.name:
            continue

        steps = get_job_steps(token, github_upstream, job.job_number).steps
        for step in steps:
            if args.step_name not in step.name.lower():
                continue

            for action in step.actions:
                if (args.step_status == "failed" and action.failed) or (
                    args.step_status != "failed" and action.failed is None
                ):
                    for item in get_action_output(token, action.output_url):
                        print(item.message)

    return 0


def help(args):
    parser.print_help()

    return 0


def parse_args():
    global parser
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description="""CircleCi utility

    Output subcommand:

    To obtain the output from failiing build jobs for the PR associated with
    the current checkout:

    > cci.py output

    Use the ----job-name and --step-name option to obtain output from failing
    'check' jobs:

    > cci.py output --job-name check --step-name check

    Specify a particular job name for obtain output from a single job:

    > ./scripts/cci.py output --job-name format-check --step-name check

""",
    )
    parser.add_argument("--token", help="CircleCi API token")

    command = parser.add_subparsers(dest="command")
    output_command_parser = command.add_parser("output")
    output_command_parser.add_argument("--pr")
    output_command_parser.add_argument("--job-name", default="build")
    output_command_parser.add_argument("--job-status", default="failed")
    output_command_parser.add_argument("--step-name", default="build")
    output_command_parser.add_argument("--step-status", default="failed")

    parser.set_defaults(command="help")

    return parser.parse_args()


def main():
    args = parse_args()
    return globals()[args.command](args)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

from typing import Iterator

from velox.py.vector import Vector


class LocalRunner:
    def __init__(self, PlanNode) -> None: ...
    def execute(self) -> Iterator[Vector]: ...
    def add_file_split(self, plan_id: str, file_path: str) -> None: ...
    def print_plan_with_stats(self) -> str: ...

def register_hive(str) -> None: ...
def register_tpch(str) -> None: ...
def unregister(str) -> None: ...



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from itertools import groupby
import json
import regex
import sys

import util

CODE_CHECKS = """*
    -abseil-*
    -android-*
    -cert-err58-cpp
    -clang-analyzer-osx-*
    -cppcoreguidelines-avoid-c-arrays
    -cppcoreguidelines-avoid-magic-numbers
    -cppcoreguidelines-pro-bounds-array-to-pointer-decay
    -cppcoreguidelines-pro-bounds-pointer-arithmetic
    -cppcoreguidelines-pro-type-reinterpret-cast
    -cppcoreguidelines-pro-type-vararg
    -fuchsia-*
    -google-*
    -hicpp-avoid-c-arrays
    -hicpp-deprecated-headers
    -hicpp-no-array-decay
    -hicpp-use-equals-default
    -hicpp-vararg
    -llvmlibc-*
    -llvm-header-guard
    -llvm-include-order
    -mpi-*
    -misc-non-private-member-variables-in-classes
    -misc-no-recursion
    -misc-unused-parameters
    -modernize-avoid-c-arrays
    -modernize-deprecated-headers
    -modernize-use-nodiscard
    -modernize-use-trailing-return-type
    -objc-*
    -openmp-*
    -readability-avoid-const-params-in-decls
    -readability-convert-member-functions-to-static
    -readability-magic-numbers
    -zircon-*
"""

# Additional opt-outs because googletest macros trip too many things.
#
TEST_CHECKS = (
    CODE_CHECKS
    + """
    -cert-err58-cpp
    -cppcoreguidelines-avoid-goto
    -cppcoreguidelines-avoid-non-const-global-variables
    -cppcoreguidelines-owning-memory
    -cppcoreguidelines-pro-type-vararg
    -cppcoreguidelines-special-member-functions
    -hicpp-avoid-goto
    -hicpp-special-member-functions
    -hicpp-vararg
    -misc-no-recursion
    -readability-implicit-bool-conversion
"""
)


def check_list(check_string):
    return ",".join([check.strip() for check in check_string.strip().splitlines()])


CODE_CHECKS = check_list(CODE_CHECKS)
TEST_CHECKS = check_list(TEST_CHECKS)


class Multimap(dict):
    def __setitem__(self, key, value):
        if key not in self:
            dict.__setitem__(self, key, [value])  # call super method to avoid recursion
        else:
            self[key].append(value)


def git_changed_lines(commit):
    file = ""
    changed_lines = Multimap()

    for line in util.run(f"git diff --text --unified=0 {commit}")[1].splitlines():
        line = line.rstrip("\n")
        fields = line.split()

        match = regex.match(r"^\+\+\+ b/.*", line)
        if match:
            file = ""

        match = regex.match(r"^\+\+\+ b/(.*(\.cpp|\.h))$", line)
        if match:
            file = match.group(1)

        match = regex.match(r"^@@", line)
        if match and file != "":
            lspan = fields[2].split(",")
            if len(lspan) <= 1:
                lspan.append(0)

            changed_lines[file] = [int(lspan[0]), int(lspan[0]) + int(lspan[1])]

    return json.dumps(
        [{"name": key, "lines": value} for key, value in changed_lines.items()]
    )


def checks(args):
    status, stdout, stderr = util.run(
        f"clang-tidy -checks='{CODE_CHECKS}' --list-checks"
    )
    print(stdout)


def check_output(output):
    return regex.match(r"^/.* warning: ", output)


def tidy(args):
    files = util.input_files(args.files)

    groups = Multimap()

    for file in files:
        groups["test" if "/tests/" in file else "main"] = file

    fix = "--fix" if args.fix == "fix" else ""
    lines = (
        ("'--line-filter=" + git_changed_lines(args.commit)) + "'"
        if args.commit is not None
        else ""
    )

    ok = True
    if groups.get("main", None):
        status, stdout, stderr = util.run(
            f"xargs clang-tidy -p=build/release/ --format-style=file -header-filter='.*' --checks='{CODE_CHECKS}' --quiet {fix} {lines}",
            input=groups["main"],
        )
        ok = check_output(stdout) and ok

    if groups.get("test", None):
        status, stdout, stderr = util.run(
            f"xargs clang-tidy -p=build/release/ --format-style=file -header-filter='.*' --checks='{TEST_CHECKS}' --quiet {fix} {lines}",
            input=groups["test"],
        )
        ok = check_output(stdout) and ok

    return 0 if ok else 1


def parse_args():
    global parser
    parser = argparse.ArgumentParser(description="CircliCi Utility")
    parser.add_argument("--commit")
    parser.add_argument("--fix")

    parser.add_argument("files", metavar="FILES", nargs="+", help="files to process")

    return parser.parse_args()


def main():
    return tidy(parse_args())


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import sys
from string import Template

cpp_template = Template(
    """\
/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// This file is generated. Do not modify it manually. To re-generate it, run:
//
//  ./velox/type/tz/gen_timezone_links.py -f /tmp/backward \\
//       > velox/type/tz/TimeZoneLinks.cpp
//
// The backward file should be the same one used in JODA,
// The latest is available here :
// https://github.com/JodaOrg/global-tz/blob/global-tz/backward.
// Presto Java currently uses the one found here:
// https://github.com/JodaOrg/global-tz/blob/2024agtz/backward
// To find the current version in Presto Java, check this file at the version of
// JODA Presto Java is using:
// https://github.com/JodaOrg/joda-time/blob/v2.12.7/src/changes/changes.xml
// @generated

#include <string>
#include <unordered_map>

namespace facebook::velox::tz {

const std::unordered_map<std::string, std::string>& getTimeZoneLinks() {
  static auto* tzLinks = new std::unordered_map<std::string, std::string>([] {
    // Work around clang compiler bug causing multi-hour compilation
    // with -fsanitize=fuzzer
    // https://github.com/llvm/llvm-project/issues/75666
    return std::unordered_map<std::string, std::string>{
$entries
    };
  }());
  return *tzLinks;
}

} // namespace facebook::velox::tz\
"""
)

entry_template = Template('        {"$tz_key", "$tz_value"},')


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="Reads an input file (specified by -f) containing mappings from "
        "time zone names to the names that should replace them and prints the "
        "corresponding `TimeZoneLinks.cpp` file.",
        epilog="(c) Facebook 2004-present",
    )
    parser.add_argument(
        "-f",
        "--file-input",
        required=True,
        help="Input timezone links file.",
    )
    return parser.parse_args()


def main():
    args = parse_arguments()
    entries = []

    # Read an input file with links following the format:
    #
    # >  Link	Australia/Sydney	Australia/NSW
    # >  Link	Australia/Darwin	Australia/North
    # >  Link	Australia/Brisbane	Australia/Queensland
    # >  ...
    #
    # Possible containing comments (everything in a line after # is ignored)

    with open(args.file_input) as file_in:
        for line in file_in:
            line = line.strip()

            columns = line.split()
            if len(columns) == 0:
                continue

            if columns[0] != "Link":
                continue

            # Officially the GMT/UTC time zone IDs start with Etc/ so Joda
            # includes this in its link files. Joda ends up removing this
            # prefix at runtime though, so we just skip ahead and do it here.
            tz_value = columns[1]
            if tz_value == "Etc/GMT" or tz_value == "Etc/UTC":
                tz_value = "UTC"

            entries.append(
                entry_template.substitute(tz_key=columns[2], tz_value=tz_value)
            )

    print(cpp_template.substitute(entries="\n".join(entries)))
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
import argparse
import fnmatch
import os
import regex
import sys

class attrdict(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__

def parse_args():
    parser = argparse.ArgumentParser(description='Update license headers')
    parser.add_argument('--header', default='license.header', help='header file')
    parser.add_argument('--extra', default=30, 
        help='extra characters past beginning of file to look for header')
    parser.add_argument('--editdist', default=7, help='max edit distance between headers')
    parser.add_argument('--remove', default=False, action="store_true", 
        help='remove the header')
    parser.add_argument('--cslash', default=False, action="store_true", 
        help='use C slash "//" style comments')
    parser.add_argument('-v', default=False, action="store_true", dest="verbose", 
        help='verbose output')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('-k', default=False, action="store_true", dest="check", 
        help='check headers')
    group.add_argument('-i', default=False, action="store_true", dest="inplace", 
        help='edit file inplace')

    parser.add_argument('files', metavar='FILES', nargs='+',
        help='files to process')

    return parser.parse_args()




#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import re
from pprint import pprint
from typing import Optional, Tuple

import benchalerts.pipeline_steps as steps
import requests
from benchalerts import AlertPipeline
from benchclients import log

DESCRIPTION = """
Analyze benchmark runs, post a GitHub Check whether there were regressions or not, and
if it's a merge-commit, post a comment back to the PR that merged the commit in.

Required environment variables:
    GITHUB_REPOSITORY - the default GITHUB_REPOSITORY given in GitHub Actions
    GITHUB_APP_ID - the ID of a GitHub App installed to this repo
    GITHUB_APP_PRIVATE_KEY - the private key file contents of the GitHub App
    CONBENCH_URL - the URL to the Conbench server where benchmark results are stored
"""


def parse_args() -> Tuple[str, str, str, float]:
    parser = argparse.ArgumentParser(
        description=DESCRIPTION, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        "--contender-sha",
        type=str,
        required=True,
        help="The SHA hash of the contender commit. The script will download benchmark "
        "information about this commit and post a GitHub Check of the full report to "
        "this commit.",
    )
    parser.add_argument(
        "--merge-commit-message",
        type=str,
        required=True,
        help="If the contender commit is a merge-commit, its message. If not, an empty "
        "string.",
    )
    parser.add_argument(
        "--z-score-threshold",
        type=float,
        required=True,
        help="The (positive) z-score threshold. Benchmarks with a z-score more extreme "
        "than this threshold will be marked as regressions.",
    )
    args = parser.parse_args()

    repo = os.environ["GITHUB_REPOSITORY"]

    return (
        repo,
        args.contender_sha,
        args.merge_commit_message,
        args.z_score_threshold,
    )


def _merge_commit_pr_number(commit_message: str) -> Optional[int]:
    """If this is a merge-commit run, get the number of the PR that was merged by
    grepping the commit message.

    Other alternatives to this strategy:

    - Hit /repos/{owner}/{repo}/commits/{commit_sha}/pulls to get the number. This
      doesn't work because of the way facebook-github-bot merges PRs.
    - Use the GitHub GraphQL API to search through recent "PR closed" events until you
      find the one that's associated with this commit. This does work but actually feels
      *more* likely to break than grepping.
    """
    if not commit_message:
        print("Merge-commit message not given; this must be a PR commit")
        return None

    # Look for e.g. (#123) at the end of the first line
    res = re.search(r"\(#(\d+)\)$", commit_message.split("\n")[0])

    if not res:
        print("Could not find the PR number in the following merge-commit message:")
        print(commit_message)
        return None

    print(f"Found a PR number (in the given merge-commit message): {repr(res[1])}")
    return int(res[1])


def main(
    repo: str,
    contender_sha: str,
    merge_commit_message: str,
    z_score_threshold: float,
):
    log.setLevel("DEBUG")

    if merge_commit_message:
        # Compare against the parent commit of the merge-commit
        baseline_run_type = steps.BaselineRunCandidates.parent
    else:
        # Compare against the default-branch commit from which the PR was forked
        baseline_run_type = steps.BaselineRunCandidates.fork_point

    pipeline_steps = [
        steps.GetConbenchZComparisonStep(
            commit_hash=contender_sha,
            baseline_run_type=baseline_run_type,
            z_score_threshold=z_score_threshold,
            step_name="z_comparison",
        ),
        steps.GitHubCheckStep(
            commit_hash=contender_sha, comparison_step_name="z_comparison", repo=repo
        ),
    ]

    # Only post a comment on merge-commits
    if merge_commit_message:
        pr_number = _merge_commit_pr_number(merge_commit_message)
        pipeline_steps.append(
            steps.GitHubPRCommentAboutCheckStep(pr_number=pr_number, repo=repo)
        )

    pipeline = AlertPipeline(steps=pipeline_steps)
    pprint(pipeline.run_pipeline())


def test_grepping():
    """This is never called. Use for local dev."""
    _merge_commit_pr_number("")
    for commit in requests.get(
        "https://api.github.com/repos/facebookincubator/velox/commits"
    ).json():
        print(commit["sha"][:7])
        _merge_commit_pr_number(commit["commit"]["message"])


if __name__ == "__main__":
    repo, contender_sha, merge_commit_message, z_score_threshold = parse_args()
    main(
        repo=repo,
        contender_sha=contender_sha,
        merge_commit_message=merge_commit_message,
        z_score_threshold=z_score_threshold,
    )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2024 by Rivos Inc.
# Licensed under the Apache License, Version 2.0, see LICENSE for details.
# SPDX-License-Identifier: Apache-2.0

import argparse
import sys
import functools
import re
from datetime import datetime

COPYRIGHT = f"""/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * Copyright (c) {datetime.now().year} by Rivos Inc.
 * Licensed under the Apache License, Version 2.0, see LICENSE for details.
 * SPDX-License-Identifier: Apache-2.0
 */
"""

parser = argparse.ArgumentParser()
parser.add_argument("--template", required=True)
parser.add_argument("-l", "--llvm-root", default=None)
parser.add_argument("--out", required=True)
parser.add_argument(
    "--backend",
    required=True,
    choices=["cuda", "hip", "sycl", "openmp", "opencl", "metal"],
)
args = parser.parse_args()

# If llvm-root not specified rely on the environment to provide libclang.
# This is typically only the case for nvidia-cuda and metal builds.
if args.llvm_root is None:
    libclang_path = None
    libclang_python = None
    clang_format = "clang-format"
else:
    libclang_path = f"{args.llvm_root}/lib"
    libclang_python = f"{args.llvm_root}/lib/python3/site-packages"
    clang_format = f"{args.llvm_root}/bin/clang-format"

# Add libclang python bindings to the library search path
if libclang_python:
    sys.path.append(libclang_python)

import clang.cindex
from clang.cindex import CursorKind

if libclang_path:
    clang.cindex.Config.set_library_path(libclang_path)

# The following options are exposed in python already, but we re-expose
# them for consistency
TU_FLAGS_None = 0x0
TU_FLAGS_DetailedPreprocessingRecord = 0x01
TU_FLAGS_Incomplete = 0x02
TU_FLAGS_PrecompiledPreamble = 0x04
TU_FLAGS_CacheCompletionResults = 0x08
TU_FLAGS_SkipFunctionBodies = 0x40
TU_FLAGS_IncludeBriefCommentsInCodeCompletion = 0x80
# Consistency checks, otherwise the newer options may also be suspect...
assert TU_FLAGS_None == clang.cindex.TranslationUnit.PARSE_NONE
assert (
    TU_FLAGS_DetailedPreprocessingRecord
    == clang.cindex.TranslationUnit.PARSE_DETAILED_PROCESSING_RECORD
)
assert TU_FLAGS_Incomplete == clang.cindex.TranslationUnit.PARSE_INCOMPLETE
assert (
    TU_FLAGS_PrecompiledPreamble
    == clang.cindex.TranslationUnit.PARSE_PRECOMPILED_PREAMBLE
)
assert (
    TU_FLAGS_CacheCompletionResults
    == clang.cindex.TranslationUnit.PARSE_CACHE_COMPLETION_RESULTS
)
assert (
    TU_FLAGS_SkipFunctionBodies
    == clang.cindex.TranslationUnit.PARSE_SKIP_FUNCTION_BODIES
)
assert (
    TU_FLAGS_IncludeBriefCommentsInCodeCompletion
    == clang.cindex.TranslationUnit.PARSE_INCLUDE_BRIEF_COMMENTS_IN_CODE_COMPLETION
)
# The following parser options are not exposed in the python bindings
# but *are* supported by libclang.
TU_FLAGS_ForSerialization = 0x10
TU_FLAGS_CXXChainedPCH = 0x20
TU_FLAGS_CreatePreambleOnFirstParse = 0x100
TU_FLAGS_KeepGoing = 0x200
TU_FLAGS_SingleFileParse = 0x400
TU_FLAGS_LimitSkipFunctionBodiesToPreamble = 0x800
TU_FLAGS_IncludeAttributedTypes = 0x1000
TU_FLAGS_VisitImplicitAttributes = 0x2000
TU_FLAGS_IgnoreNonErrorsFromIncludedFiles = 0x4000
TU_FLAGS_RetainExcludedConditionalBlocks = 0x8000


def parse(filename):
    index = clang.cindex.Index.create()
    tu = index.parse(
        filename,
        args=["-xc++", "-std=c++17"],
        options=clang.cindex.TranslationUnit.PARSE_DETAILED_PROCESSING_RECORD
        | clang.cindex.TranslationUnit.PARSE_INCOMPLETE
        | TU_FLAGS_KeepGoing
        | TU_FLAGS_SingleFileParse
        | TU_FLAGS_IncludeAttributedTypes
        | TU_FLAGS_VisitImplicitAttributes
        | TU_FLAGS_IgnoreNonErrorsFromIncludedFiles
        | TU_FLAGS_RetainExcludedConditionalBlocks,
    )
    return index, tu


@functools.lru_cache
def read_cached_file(path):
    f = open(path, "r")
    return f.read()


def read_extent(extent):
    file = extent.start.file
    return read_cached_file(file.name)[extent.start.offset : extent.end.offset]


FUNCTION_CURSOR_KINDS = [
    CursorKind.FUNCTION_DECL,
    CursorKind.CXX_METHOD,
    CursorKind.FUNCTION_TEMPLATE,
]
TEMPLATE_CURSOR_KINDS = [
    CursorKind.TEMPLATE_TYPE_PARAMETER,
    CursorKind.TEMPLATE_NON_TYPE_PARAMETER,
    CursorKind.TEMPLATE_TEMPLATE_PARAMETER,
]


# Helper for extracting commonly needed properties from a cursor
def extract_node(cursor):
    node = {
        "kind": cursor.kind,
        "displayname": cursor.displayname,
        "spelling": cursor.spelling,
        "cursor": cursor,
    }

    if cursor.type.spelling != "":
        node["type_spelling"] = cursor.type.spelling

    if cursor.kind in FUNCTION_CURSOR_KINDS:
        template_params = []
        function_params = []
        raw_attrs = []
        body = None
        for c in cursor.get_children():
            if c.kind == CursorKind.PARM_DECL:
                function_params.append(c)
            elif c.kind in TEMPLATE_CURSOR_KINDS:
                template_params.append(c)
            # Since unknown attributes are dropped during parse we
            # encode them using clang::annotate.
            elif c.kind == CursorKind.ANNOTATE_ATTR:
                raw_attrs.append(c)
            elif c.kind == CursorKind.COMPOUND_STMT:
                if body is not None:
                    raise ValueError("Multiple compound statements unexpected")
                body = c
            else:
                raise ValueError(
                    f"Unhandled method child, \
                        pessimistically exiting: {c.kind} {c.extent}"
                )

        attrs = {}
        for c in raw_attrs:
            a = c.spelling.split("=", 1)
            attrs.setdefault(a[0], []).append(a[1] if len(a) == 2 else True)

        node["fn_props"] = {
            "attrs": attrs,
            "return_type": cursor.result_type,
            "template_params": template_params,
            "function_params": function_params,
            "body": body,
        }

    node["children"] = list(cursor.get_children())

    return node


snake_caser = re.compile(r"(?<!^)([A-Z])")


def to_snake_case(name):
    return snake_caser.sub(r"_\1", name).lower()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

import unittest
from velox.py.type import (
    Type,
    BIGINT,
    INTEGER,
    SMALLINT,
    TINYINT,
    BOOLEAN,
    REAL,
    DOUBLE,
    VARCHAR,
    VARBINARY,
    ARRAY,
    MAP,
    ROW,
)


class TestPyVeloxTypes(unittest.TestCase):
    def test_simple_types(self):
        self.assertTrue(isinstance(BIGINT(), Type))
        self.assertTrue(isinstance(INTEGER(), Type))
        self.assertTrue(isinstance(SMALLINT(), Type))
        self.assertTrue(isinstance(TINYINT(), Type))
        self.assertTrue(isinstance(BOOLEAN(), Type))
        self.assertTrue(isinstance(REAL(), Type))
        self.assertTrue(isinstance(DOUBLE(), Type))
        self.assertTrue(isinstance(VARCHAR(), Type))
        self.assertTrue(isinstance(VARBINARY(), Type))

    def test_complex_types(self):
        self.assertTrue(isinstance(ARRAY(VARCHAR()), Type))
        self.assertTrue(isinstance(MAP(BIGINT(), VARCHAR()), Type))
        self.assertTrue(
            isinstance(ROW(["c0", "c1"], [INTEGER(), ARRAY(VARCHAR())]), Type)
        )
        self.assertTrue(isinstance(ROW(), Type))

        # Invalid complex types.
        self.assertRaises(TypeError, ARRAY)
        self.assertRaises(TypeError, MAP)
        self.assertRaises(TypeError, MAP, BIGINT())
        self.assertRaises(RuntimeError, ROW, ["col1"])

    def test_to_str(self):
        self.assertEqual("BOOLEAN", str(BOOLEAN()))
        self.assertEqual("VARBINARY", str(VARBINARY()))

        self.assertEqual("ARRAY<DOUBLE>", str(ARRAY(DOUBLE())))
        self.assertEqual("MAP<VARCHAR,DOUBLE>", str(MAP(VARCHAR(), DOUBLE())))
        self.assertEqual(
            "ROW<c0:INTEGER,c1:TINYINT>", str(ROW(["c0", "c1"], [INTEGER(), TINYINT()]))
        )
        velox_type = ROW(
            ["c0", "c1"],
            [MAP(MAP(ARRAY(ARRAY(REAL())), DOUBLE()), BOOLEAN()), TINYINT()],
        )
        self.assertEqual(
            "ROW<c0:MAP<MAP<ARRAY<ARRAY<REAL>>,DOUBLE>,BOOLEAN>,c1:TINYINT>",
            str(velox_type),
        )

    def test_equality(self):
        self.assertEqual(INTEGER(), INTEGER())
        self.assertEqual(
            MAP(MAP(TINYINT(), BOOLEAN()), BOOLEAN()),
            MAP(MAP(TINYINT(), BOOLEAN()), BOOLEAN()),
        )
        self.assertEqual(ROW(["a"], [INTEGER()]), ROW(["a"], [INTEGER()]))

        self.assertNotEqual(BIGINT(), INTEGER())
        self.assertNotEqual(ARRAY(BIGINT()), REAL())



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import os
import pathlib
import re
import subprocess
import sys
import tempfile

from collections import defaultdict


_OUTPUT_NUM_COLS = 100


# Cosmetic helper functions.
# GitHub Actions does not provide a tty but can still display colors
def in_gha() -> bool:
    return os.getenv("GITHUB_ACTIONS") == "true"


def color_red(text) -> str:
    return (
        "\033[91m{}\033[00m".format(text) if sys.stdout.isatty() or in_gha() else text
    )


def color_yellow(text) -> str:
    return (
        "\033[93m{}\033[00m".format(text) if sys.stdout.isatty() or in_gha() else text
    )


def color_green(text) -> str:
    return (
        "\033[92m{}\033[00m".format(text) if sys.stdout.isatty() or in_gha() else text
    )


def bold(text) -> str:
    return "\033[1m{}\033[00m".format(text) if sys.stdout.isatty() or in_gha() else text


def get_benchmark_handle(file_path, name):
    if name[0] == "%":
        name = name[1:]
    return "{}/{}".format(os.path.basename(file_path), name)


def fmt_runtime(time_ns):
    if time_ns < 1000:
        return "{:.2f}ns".format(time_ns)

    time_usec = time_ns / 1000
    if time_usec < 1000:
        return "{:.2f}us".format(time_usec)
    else:
        return "{:.2f}ms".format(time_usec / 1000)


def get_retry_name(args, file_name):
    """
    Extract the subdir name between the base path and file name, and use that
    as the retry name. Top level files will have retry name set to "."
    """
    path = _normalize_path(file_name)
    try:
        parent_path = path.relative_to(_normalize_path(args.contender_path))
    except:
        parent_path = path.relative_to(_normalize_path(args.baseline_path))
    return str(parent_path.parent)


def compare_file(args, target_data, baseline_data):
    def preprocess_data(input_map):
        output_map = defaultdict(dict)
        for file_name, data in input_map.items():
            retry = get_retry_name(args, file_name)
            for row in data:
                # Folly benchmark exports line separators by mistake as an entry on
                # the json file.
                if row[1] == "-":
                    continue
                output_map[(row[0], row[1])][retry] = row[2]
        return output_map

    baseline_map = preprocess_data(baseline_data)
    target_map = preprocess_data(target_data)

    passes = []
    faster = []
    failures = []

    # Iterate over each benchmark.
    for handle, target_values in target_map.items():
        retries = len(baseline_map[handle])
        if retries != len(target_values):
            print("Wrong number of retries found. Skipping '{}'".format(handle))
            continue

        # Iterate over each retry. The first iteration is always the full
        # execution, then optionally a few retries.
        for retry, target_result in sorted(target_values.items()):
            is_last = retry == sorted(target_values.keys())[-1]
            is_first = retry == "."
            baseline_result = baseline_map[handle][retry]

            # Calculate delta between baseline and target results.
            if baseline_result == 0 or target_result == 0:
                delta = 0
            elif baseline_result > target_result:
                delta = 1 - (target_result / baseline_result)
            else:
                delta = (1 - (baseline_result / target_result)) * -1

            # Set status message based on the delta and number of retries.
            if not is_last:
                if delta > 0:
                    status = color_yellow("🗲 Redo")
                else:
                    status = color_yellow("✗ Redo")

            # If there are no more retries and this exceeded the threshold.
            elif abs(delta) > args.threshold:
                if delta > 0:
                    status = color_green("🗲 Pass")
                    passes.append((handle[0], handle[1], delta))
                    faster.append((handle[0], handle[1], delta))
                else:
                    status = color_red("✗ Fail")
                    failures.append((handle[0], handle[1], delta))
            else:
                status = color_green("✓ Pass")
                passes.append((handle[0], handle[1], delta))

            suffix = "({} vs {}) {:+.2f}%".format(
                fmt_runtime(baseline_result), fmt_runtime(target_result), delta * 100
            )
            bm_handle = get_benchmark_handle(*handle)

            # Add retry information.
            if not is_first:
                bm_handle += " ({})".format(retry)
                status = "  " + status

            spacing = " " * (
                _OUTPUT_NUM_COLS - (len(status) + len(bm_handle) + len(suffix) - 4)
            )
            print("    {}: {}{}{}".format(status, bm_handle, spacing, suffix))

    return passes, faster, failures


def find_json_files(path: pathlib.Path, recursive=False):
    """Finds json files in a given directory. Supports recursive searchs."""
    pattern = "*.json"
    files = path.rglob(pattern) if recursive else path.glob(pattern)
    json_files = defaultdict(list)

    for path in files:
        json_files[path.name] += [path.resolve()]
    return json_files


def read_json_files(file_names):
    """Reads a sequence of json files and returns their contents in a dict: {file_name: content}"""
    output_map = {}
    for file_name in file_names:
        with open(file_name) as f:
            output_map[file_name] = json.load(f)
    return output_map


def compare(args):
    print(
        "=> Starting comparison using {} ({}%) as threshold.".format(
            args.threshold, args.threshold * 100
        )
    )
    print("=> Values are reported as percentage normalized to the largest values:")
    print("=>    (positive means speedup; negative means regression).")

    # Read file lists from both directories.
    baseline_map = find_json_files(pathlib.Path(args.baseline_path), args.recursive)
    target_map = find_json_files(pathlib.Path(args.contender_path), args.recursive)

    all_passes = []
    all_faster = []
    all_failures = []

    # Keep track of benchmarks that exceeded the threshold to they can be saved
    # to the rerun_output file.
    rerun_log = {}

    # Compare json results from each file.
    for file_name, contender_path in target_map.items():
        print("=" * _OUTPUT_NUM_COLS)
        print(file_name)
        print("=" * _OUTPUT_NUM_COLS)

        if file_name not in baseline_map:
            print("WARNING: baseline file for '%s' not found. Skipping." % file_name)
            continue

        target_data = read_json_files(contender_path)
        baseline_data = read_json_files(baseline_map[file_name])

        passes, faster, failures = compare_file(args, target_data, baseline_data)
        all_passes += passes
        all_faster += faster
        all_failures += failures

        if faster or failures:
            rerun_log[file_name] = faster + failures

    def print_list(names):
        for n in names:
            print(
                "    {} ({:+.2f}%)".format(get_benchmark_handle(n[0], n[1]), n[2] * 100)
            )

    # Write rerun log to output file.
    if args.rerun_json_output:
        with open(args.rerun_json_output, "w") as out_file:
            if rerun_log:
                out_file.write(json.dumps(rerun_log, indent=4))

    # Print a nice summary of the results:
    print("Summary ({}% threshold):".format(args.threshold * 100))
    if all_passes:
        faster_summary = (
            " ({} are faster):".format(len(all_faster)) if all_faster else ""
        )
        print(color_green("  Pass: %d%s " % (len(all_passes), faster_summary)))
        print_list(all_faster)

    if all_failures:
        print(color_red("  Fail: %d" % len(all_failures)))
        print_list(all_failures)
        if not args.do_not_fail:
            return 1
    return 0


def _find_binaries(binary_path: pathlib.Path):
    print(f"Looking for binaries at '{binary_path}'")

    # Must run `make benchmarks-basic-build` before this
    binaries = [
        path
        for path in binary_path.glob("*")
        if os.access(path, os.X_OK) and path.is_file()
    ]
    if not binaries:
        raise ValueError(f"No binaries found at path '{binary_path.resolve()}'")

    print(f"Found {len(binaries)} benchmark binaries")
    return binaries


def _default_binary_path():
    repo_root = pathlib.Path(__file__).parent.parent.absolute()
    return repo_root.joinpath("_build", "release", "velox", "benchmarks", "basic")


def _normalize_path(binary_path: str) -> pathlib.Path:
    path = pathlib.Path(binary_path)
    if not path.is_absolute():
        path = pathlib.Path.cwd().joinpath(path).resolve()
    return path


def run_all_benchmarks(
    output_dir,
    binary_path=None,
    binary_filter=None,
    bm_filter=None,
    bm_max_secs=None,
    bm_max_trials=None,
    bm_estimate_time=False,
):
    if binary_path:
        binary_path = _normalize_path(binary_path)
    else:
        binary_path = _default_binary_path()

    binaries = _find_binaries(binary_path)
    output_dir_path = pathlib.Path(output_dir)
    output_dir_path.mkdir(parents=True, exist_ok=True)

    for binary_path in binaries:
        if binary_filter and not re.search(binary_filter, binary_path.name):
            continue

        out_path = output_dir_path / f"{binary_path.name}.json"
        print(f"Executing and dumping results for '{binary_path}' to '{out_path}':")
        run_command = [
            binary_path,
            "--bm_json_verbose",
            out_path,
        ]

        if bm_max_secs:
            run_command.extend(["--bm_max_secs", str(bm_max_secs)])

        if bm_max_trials:
            run_command.extend(["--bm_max_trials", str(bm_max_trials)])

        if bm_filter:
            run_command.extend(["--bm_regex", bm_filter])

        if bm_estimate_time:
            run_command.append("--bm_estimate_time")

        try:
            print(run_command)
            subprocess.run(run_command, check=True)
        except subprocess.CalledProcessError as e:
            print(e.stderr.decode("utf-8"))
            raise e


def upload_results(args):
    output_dir = args.output_dir
    run_id = args.run_id
    sha = args.sha
    pr_number = args.pr_number

    print(f"Uploading results from {output_dir=} to Conbench with {run_id=}")

    # Check there's actually results first
    if not list(pathlib.Path(output_dir).rglob("*.json")):
        print("No results to upload")
        return

    # Import benchadapt inside this function so you don't need it if you're just running benchmarks
    from benchadapt.adapters import FollyAdapter
    from benchadapt.log import log

    # Print the POSTs in the logs
    log.setLevel("DEBUG")

    # benchadapt needs these for logging in to Conbench
    required_env_vars = {"CONBENCH_URL", "CONBENCH_EMAIL", "CONBENCH_PASSWORD"}
    missing_env_vars = required_env_vars - set(os.environ)
    if missing_env_vars:
        print(
            "Not uploading results to Conbench because these env vars are missing: "
            f"{missing_env_vars}"
        )
        return

    pr_number = int(pr_number) if pr_number else None
    run_reason = "pull request" if pr_number else "commit"
    run_name = f"{run_reason}: {sha}"

    conbench_upload_callable = FollyAdapter(
        # Since benchmarks have already run, this run command is a no-op
        command=["ls", output_dir],
        result_dir=output_dir,
        result_fields_override={
            "run_id": run_id,
            "run_name": run_name,
            "run_reason": run_reason,
            "github": {
                "repository": "https://github.com/facebookincubator/velox",
                "pr_number": pr_number,
                "commit": sha,
            },
        },
        result_fields_append={
            "info": {
                # TODO: undo the hard coding with ways to get the values from velox
                # c.f. https://github.com/ursacomputing/benchmarks/blob/033eee0951adbf41931a2de95caccbac887da6ff/benchmarks/_benchmark.py#L30-L48
                "velox_version": "0.0.1",
                "velox_compiler_id": None,
                "velox_compiler_version": None,
            },
            "context": {"velox_compiler_flags": None},
        },
    )

    conbench_upload_callable()


def run(args):
    output_dir = args.output_path or tempfile.mkdtemp()
    kwargs = {
        "output_dir": output_dir,
        "binary_path": args.binary_path,
        "binary_filter": args.binary_filter,
        "bm_filter": args.bm_filter,
        "bm_max_secs": args.bm_max_secs,
        "bm_max_trials": args.bm_max_trials,
        "bm_estimate_time": args.bm_estimate_time,
    }

    # In case we only want to rerun failed benchmarks from rerun_json_input.
    if args.rerun_json_input:
        with open(args.rerun_json_input, "r") as input_file:
            json_input = json.load(input_file)

        def gen_binary_filter(json_file_name):
            return "^{}$".format(json_file_name.rstrip(".json"))

        def gen_bm_filter(bm_list):
            return "^{}$".format("|".join([x[1] for x in bm_list]))

        for file_name, bm_list in json_input.items():
            kwargs["binary_filter"] = gen_binary_filter(file_name)
            kwargs["bm_filter"] = gen_bm_filter(bm_list)
            run_all_benchmarks(**kwargs)

    # Otherwise, run all benchmarks we can find.
    else:
        run_all_benchmarks(**kwargs)


def parse_args():
    parser = argparse.ArgumentParser(description="Velox Benchmark Runner Utility.")
    parser.set_defaults(func=lambda _: parser.print_help())

    subparsers = parser.add_subparsers(help="Please specify one of the subcommands.")

    # Arguments for the "run" subparser.
    parser_run = subparsers.add_parser("run", help="Run benchmarks and dump results.")
    parser_run.set_defaults(func=run)
    parser_run.add_argument(
        "--binary_path",
        default=None,
        help="Directory where benchmark binaries are stored. "
        "Defaults to release build directory.",
    )
    parser_run.add_argument(
        "--output_path",
        default=None,
        help="Directory where output json files will be written to. "
        "By default generate a temporary directory.",
    )
    parser_run.add_argument(
        "--binary_filter",
        default=None,
        help="Filter applied to binary names. "
        "By default execute all binaries found.",
    )
    parser_run.add_argument(
        "--bm_filter",
        default=None,
        help="Filter applied to benchmark names within binaries. "
        "By default execute all benchmarks.",
    )
    parser_run.add_argument(
        "--bm_max_secs",
        default=None,
        type=int,
        help="For how many seconds to run each benchmark in a binary.",
    )
    parser_run.add_argument(
        "--bm_max_trials",
        default=None,
        type=int,
        help="Maximum number of trials (iterations) executed for each benchmark.",
    )
    parser_run.add_argument(
        "--bm_estimate_time",
        default=False,
        action="store_true",
        help="Use folly benchmark --bm_estimate_time flag.",
    )
    parser_run.add_argument(
        "--rerun_json_input",
        default=None,
        help="Only binaries and benchmark names read from this json file will "
        "be run. This file needs to be generated using the "
        "--rerun_json_output flag.",
    )

    # Arguments for the "compare" subparser.
    parser_compare = subparsers.add_parser(
        "compare", help="Compare benchmark dumped results."
    )
    parser_compare.set_defaults(func=compare)
    parser_compare.add_argument(
        "--baseline_path",
        required=True,
        help="Path where containing base dump results.",
    )
    parser_compare.add_argument(
        "--contender_path",
        required=True,
        help="Path where containing target dump results.",
    )
    parser_compare.add_argument(
        "-t",
        "--threshold",
        type=float,
        default=0.05,
        help="Comparison threshold. "
        "Variations larger than this threshold will be reported as failures. "
        "Default 0.05 (5%%).",
    )
    parser_compare.add_argument(
        "--rerun_json_output",
        default=None,
        help="File where the rerun output will be saved. Redo output contains "
        "information about the failed benchmarks (the ones where the variation "
        "exceeded the threshold).",
    )
    parser_compare.add_argument(
        "--recursive",
        default=False,
        action="store_true",
        help="Looks for json files recursively, understanding subdirs as "
        "retries and printing the output accordingly.",
    )
    parser_compare.add_argument(
        "--do_not_fail",
        default=False,
        action="store_true",
        help="Do not return failure code if comparisons fail.",
    )

    parser_upload = subparsers.add_parser(
        "upload",
        help="Upload benchmark results to conbench. Requires the `benchadapt` package "
        "installed and the following env vars set: "
        "CONBENCH_URL, CONBENCH_EMAIL, CONBENCH_PASSWORD",
    )
    parser_upload.set_defaults(func=upload_results)
    parser_upload.add_argument(
        "--output_dir",
        default=None,
        help="Location of the benchmark results.",
    )
    parser_upload.add_argument(
        "--run_id",
        default=None,
        help="A Conbench run ID unique to this build.",
    )
    parser_upload.add_argument(
        "--sha",
        default=None,
        help="HEAD sha for the result upload to conbench.",
    )
    parser_upload.add_argument(
        "--pr_number",
        default=None,
        help="PR number for the result upload to conbench.",
    )

    return parser.parse_args()


def main():
    args = parse_args()
    return args.func(args)


if __name__ == "__main__":
    sys.exit(main())



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gzip
import json
import os
import regex
import subprocess
import sys


class attrdict(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


class string(str):
    def extract(self, rexp):
        return regex.match(rexp, self).group(1)

    def json(self):
        return json.loads(self, object_hook=attrdict)


def run(command, compressed=False, **kwargs):
    if "input" in kwargs:
        input = kwargs["input"]

        if type(input) == list:
            input = "\n".join(input) + "\n"

        kwargs["input"] = input.encode("utf-8")

    reply = subprocess.run(
        command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs
    )

    if compressed:
        stdout = gzip.decompress(reply.stdout)
    else:
        stdout = reply.stdout

    stdout = (
        string(stdout.decode("utf-8", errors="ignore").strip())
        if stdout is not None
        else ""
    )
    stderr = (
        string(reply.stderr.decode("utf-8").strip()) if reply.stderr is not None else ""
    )

    if stderr != "":
        print(stderr, file=sys.stderr)

    return reply.returncode, stdout, stderr


def get_filename(filename):
    return os.path.basename(filename)


def get_fileextn(filename):
    split = os.path.splitext(filename)
    if len(split) <= 1:
        return ""

    return split[-1]


def script_path():
    return os.path.dirname(os.path.realpath(sys.argv[0]))


def input_files(files):
    if len(files) == 1 and files[0] == "-":
        return [file.strip() for file in sys.stdin.readlines()]
    else:
        return files



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest
import pyarrow
import tempfile

from velox.py.arrow import to_velox
from velox.py.plan_builder import PlanBuilder
from velox.py.file import DWRF
from velox.py.type import BIGINT, ROW
from velox.py.runner import LocalRunner, register_hive, register_tpch, unregister


class TestPyVeloxRunner(unittest.TestCase):
    def test_runner_empty(self):
        plan_builder = PlanBuilder().values()
        runner = LocalRunner(plan_builder.get_plan_node())
        total_size = 0

        for vector in runner.execute():
            total_size += vector.size()
        self.assertEqual(total_size, 0)

    def test_runner_not_executed(self):
        # Ensure it won't hang on destruction when it was not executed.
        plan_builder = PlanBuilder().values()
        LocalRunner(plan_builder.get_plan_node())

    def test_runner_executed_twice(self):
        # Ensure the runner fails if it is executed twice.
        plan_builder = PlanBuilder().values()
        runner = LocalRunner(plan_builder.get_plan_node())
        runner.execute()
        self.assertRaises(RuntimeError, runner.execute)

    def test_runner_with_values(self):
        vectors = []
        batch_size = 10
        num_batches = 10

        for i in range(num_batches):
            array = pyarrow.array(list(range(i * batch_size, (i + 1) * batch_size)))
            batch = pyarrow.record_batch([array], names=["c0"])
            vectors.append(to_velox(batch))

        plan_builder = PlanBuilder().values(vectors)
        runner = LocalRunner(plan_builder.get_plan_node())
        total_size = 0

        for vector in runner.execute():
            total_size += vector.size()
        self.assertEqual(total_size, 100)

    def test_runner_with_values_order_limit(self):
        vectors = []
        batch_size = 10
        num_batches = 10

        for i in range(num_batches):
            array = pyarrow.array(list(range(i * batch_size, (i + 1) * batch_size)))
            batch = pyarrow.record_batch([array], names=["c0"])
            vectors.append(to_velox(batch))

        plan_builder = (
            PlanBuilder().values(vectors).order_by(["c0 DESC"]).limit(5, offset=2)
        )
        runner = LocalRunner(plan_builder.get_plan_node())

        iterator = runner.execute()
        output = next(iterator)
        self.assertRaises(StopIteration, next, iterator)

        expected_result = to_velox(
            pyarrow.record_batch([pyarrow.array([97, 96, 95, 94, 93])], names=["c0"])
        )
        self.assertEqual(output, expected_result)

    def test_runner_with_join(self):
        batch_size = 10
        array = pyarrow.array([42] * batch_size)
        batch = to_velox(pyarrow.record_batch([array], names=["c0"]))

        plan_builder = PlanBuilder()
        plan_builder.values([batch]).merge_join(
            left_keys=["c0"],
            right_keys=["c0"],
            right_plan_node=(
                plan_builder.new_builder().values([batch]).get_plan_node()
            ),
        )

        runner = LocalRunner(plan_builder.get_plan_node())
        total_size = 0

        for vector in runner.execute():
            total_size += vector.size()
        self.assertEqual(total_size, batch_size * batch_size)

    def test_runner_with_merge_sort(self):
        array = pyarrow.array([0, 1, 2, 3, 4])
        batch = to_velox(pyarrow.record_batch([array], names=["c0"]))

        plan_builder = PlanBuilder()
        plan_builder.sorted_merge(
            keys=["c0"],
            sources=(
                plan_builder.new_builder().values([batch]).get_plan_node(),
                plan_builder.new_builder().values([batch]).get_plan_node(),
                plan_builder.new_builder().values([batch]).get_plan_node(),
            ),
        )

        runner = LocalRunner(plan_builder.get_plan_node())
        iterator = runner.execute()
        output = next(iterator)
        self.assertRaises(StopIteration, next, iterator)

        expected = to_velox(
            pyarrow.record_batch(
                [pyarrow.array([0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4])],
                names=["c0"],
            )
        )
        self.assertEqual(output, expected)

    def test_register_connectors(self):
        register_hive("conn1")
        self.assertRaises(RuntimeError, register_hive, "conn1")
        register_tpch("conn2")

        unregister("conn1")
        unregister("conn2")
        self.assertRaises(RuntimeError, unregister, "conn3")
        register_tpch("conn1")
        unregister("conn1")
        register_tpch("conn2")

    def test_write_read_file(self):
        # Test writing a batch of data to a dwrf file on disk, then
        # reading it back.
        register_hive()

        # Generate input data.
        batch_size = 10
        array = pyarrow.array([42] * batch_size)
        input_batch = to_velox(pyarrow.record_batch([array], names=["c0"]))

        with tempfile.TemporaryDirectory() as temp_dir:
            output_file = f"{temp_dir}/output_file"

            plan_builder = PlanBuilder()
            plan_builder.values([input_batch]).table_write(
                output_file=DWRF(output_file),
                connector_id="hive",
            )

            # Execute and write to output file.
            runner = LocalRunner(plan_builder.get_plan_node())
            iterator = runner.execute()
            output = next(iterator)
            self.assertRaises(StopIteration, next, iterator)
            self.assertNotEqual(runner.print_plan_with_stats(), "")

            output_file_from_table_writer = self.extract_file(output)
            self.assertEqual(output_file, output_file_from_table_writer)

            # Now scan it back.
            scan_plan_builder = PlanBuilder()
            scan_plan_builder.table_scan(
                output_schema=ROW(["c0"], [BIGINT()]),
                connector_id="hive",
                input_files=[DWRF(output_file)],
            )

            runner = LocalRunner(scan_plan_builder.get_plan_node())
            iterator = runner.execute()
            result = next(iterator)
            self.assertRaises(StopIteration, next, iterator)

            # Ensure the read batch is the same as the one written.
            self.assertEqual(input_batch, result)

    def test_tpch_gen(self):
        register_tpch("tpch")
        register_hive("hive")

        # Generate lineitem, write to an output file, then read it back.
        with tempfile.TemporaryDirectory() as temp_dir:
            output_file = f"{temp_dir}/output_file"

            plan_builder = PlanBuilder()
            plan_builder.tpch_gen(
                table_name="lineitem",
                connector_id="tpch",
                scale_factor=0.001,
            ).table_write(
                output_file=DWRF(output_file),
                connector_id="hive",
            )

            # Execute and write to output file.
            runner = LocalRunner(plan_builder.get_plan_node())
            iterator = runner.execute()
            output = next(iterator)
            self.assertRaises(StopIteration, next, iterator)

            output_file_from_table_writer = self.extract_file(output)
            self.assertEqual(output_file, output_file_from_table_writer)

            # Now scan it back.
            scan_plan_builder = PlanBuilder()
            scan_plan_builder.table_scan(
                output_schema=ROW(["l_orderkey", "l_partkey"], [BIGINT()] * 2),
                connector_id="hive",
                input_files=[DWRF(output_file)],
            )

            runner = LocalRunner(scan_plan_builder.get_plan_node())
            output_rows = 0

            for vector in runner.execute():
                output_rows += vector.size()
            self.assertEqual(output_rows, 6005)

    def extract_file(self, output_vector):
        # Parse and return the output file name from the writer's output.
        output_json = json.loads(output_vector.child_at(1)[1])
        self.assertIsNotNone(output_json)

        write_infos = output_json["fileWriteInfos"]
        self.assertTrue(isinstance(write_infos, list))
        self.assertGreater(len(write_infos), 0)
        return write_infos[0].get("targetFileName", "")



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

from enum import Enum
from typing import List, Dict, Type

from velox.py.file import File
from velox.py.type import Type


class JoinType(Enum):
    INNER = 1
    LEFT = 2
    RIGHT = 3
    FULL = 4

class PlanNode: ...
class PlanBuilder:
    def __init__(self) -> None: ...
    def table_scan(
        self,
        output_schema: Type,
        aliases: Dict[str, str] = {},
        subfields: Dict[str, List[int]] = {},
        row_index: str = "",
        connector_id: str = "prism",
        input_files: List[File] = [],
    ) -> PlanBuilder: ...
    def get_plan_node(self) -> PlanBuilder: ...
    def new_builder(self) -> PlanBuilder: ...
    def id(self) -> str: ...



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pyvelox.pyvelox as pv
import unittest


class TestVeloxExpression(unittest.TestCase):
    def test_from_string(self):
        pv.Expression.from_string("a + b")

    def test_eval(self):
        expr = pv.Expression.from_string("a + b")
        a = pv.from_list([1, 2, 3, None])
        b = pv.from_list([4, 5, None, 6])
        c = expr.evaluate(["a", "b"], [a, b])
        self.assertEqual(c[0], 5)
        self.assertEqual(c[1], 7)
        self.assertEqual(c[2], None)
        self.assertEqual(c[3], None)

        # Ignore extra arguments
        d = pv.from_list([1, 2, 3, None])
        e = expr.evaluate(["a", "b", "d"], [a, b, d])
        self.assertEqual(e[0], 5)
        self.assertEqual(e[1], 7)
        self.assertEqual(e[2], None)
        self.assertEqual(e[3], None)

        with self.assertRaises(ValueError):
            d = pv.from_list([10, 11])
            expr.evaluate(["a", "b"], [a, d])

        with self.assertRaises(RuntimeError):
            d = pv.from_list(["hi", "bye", "hello", "goodbye"])
            expr.evaluate(["a", "b"], [a, d])

    def test_eval_map(self):
        expr = pv.Expression.from_string("a + b")
        a = pv.from_list([1, 2, 3, None])
        b = pv.from_list([4, 5, None, 6])
        c = expr.evaluate({"a": a, "b": b})
        self.assertEqual(c[0], 5)
        self.assertEqual(c[1], 7)
        self.assertEqual(c[2], None)
        self.assertEqual(c[3], None)

        # Ignore extra arguments
        d = pv.from_list([1, 2, 3, None])
        e = expr.evaluate({"a": a, "b": b, "d": d})
        self.assertEqual(e[0], 5)
        self.assertEqual(e[1], 7)
        self.assertEqual(e[2], None)
        self.assertEqual(e[3], None)

    def test_functions(self):
        concat = pv.Expression.from_string("concat(a, b)")

        a = pv.from_list(["Hello", "Happy"])
        b = pv.from_list([" world", " birthday"])

        c = concat.evaluate({"a": a, "b": b})
        self.assertEqual(c[0], "Hello world")
        self.assertEqual(c[1], "Happy birthday")

    def test_cast(self):
        cast_to_int = pv.Expression.from_string("cast(a AS bigint)")

        strs = pv.from_list(["1", "10", "100"])
        ints = cast_to_int.evaluate({"a": strs})
        self.assertEqual(ints[0], 1)
        self.assertEqual(ints[1], 10)
        self.assertEqual(ints[2], 100)

    def test_all_nulls(self):
        # A vector of all None is currently not constructable, so we hack it using an expression to return Nones
        expr = pv.Expression.from_string("a + b")
        a = pv.from_list([1, None])
        b = pv.from_list([None, 2])
        c = expr.evaluate({"a": a, "b": b})
        d = expr.evaluate({"a": a, "b": b})
        self.assertEqual(c[0], None)
        self.assertEqual(c[1], None)

        self.assertEqual(d[0], None)
        self.assertEqual(d[1], None)

        e = expr.evaluate({"a": c, "b": d})
        self.assertEqual(e[0], None)
        self.assertEqual(e[1], None)

    def test_complicated_expression(self):
        expr = pv.Expression.from_string(
            "(cast(0 - b as DOUBLE) + sqrt(cast(b * b - 4 * a * c as DOUBLE))) / cast((2 * a) as double)"
        )
        a = pv.from_list([1, 3])
        b = pv.from_list([-7, 0])
        c = pv.from_list([10, -12])

        expected = [5, 2]
        actual = expr.evaluate({"a": a, "b": b, "c": c})
        for e, a in zip(expected, actual):
            self.assertEqual(e, a)

    def test_switch_case(self):
        expr = pv.Expression.from_string(
            "case a when 1 then 'one' when 2 then 'two' else 'many' end"
        )
        a = pv.from_list([1, 2, 3, 4, 5, 1])
        b = expr.evaluate({"a": a})
        self.assertEqual(b[0], "one")
        self.assertEqual(b[1], "two")
        self.assertEqual(b[2], "many")
        self.assertEqual(b[3], "many")
        self.assertEqual(b[4], "many")
        self.assertEqual(b[5], "one")



#!/usr/bin/env python3
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import sys
from string import Template

cpp_template = Template(
    """\
/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// This file is generated. Do not modify it manually. To re-generate it, run:
//
//  ./velox/type/tz/gen_timezone_database.py -f /tmp/zone-index.properties \\
//       > velox/type/tz/TimeZoneDatabase.cpp
//
// The zone-index.properties file should be the same one used in Presto,
// Available here :
// https://github.com/prestodb/presto/blob/master/presto-common/src/main/resources/com/facebook/presto/common/type/zone-index.properties.
// @generated

#include <string>
#include <unordered_map>
#include <vector>

namespace facebook::velox::util {

const std::vector<std::pair<int16_t, std::string>>& getTimeZoneEntries() {
  static auto* tzDB = new std::vector<std::pair<int16_t, std::string>>([] {
    // Work around clang compiler bug causing multi-hour compilation
    // with -fsanitize=fuzzer
    // https://github.com/llvm/llvm-project/issues/75666
    return std::vector<std::pair<int16_t, std::string>>{
$entries
    };
  }());
  return *tzDB;
}

} // namespace facebook::velox::util\
"""
)

entry_template = Template('        {$tz_id, "$tz_name"},')


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="Reads an input timezone mapping database file "
        "(specified by -f) and prints the corresponding "
        "`TimeZoneDatabase.cpp` file.",
        epilog="(c) Facebook 2004-present",
    )
    parser.add_argument(
        "-f",
        "--file-input",
        required=True,
        help="Input timezone database file (space separated, one entry per line).",
    )
    return parser.parse_args()


def main():
    args = parse_arguments()
    entries = []

    # Read an input file following the format:
    #
    # >  1825 America/Los_Angeles
    # >  1826 America/Louisville
    # >  1827 America/Lower_Princes
    # >  ...
    #

    # Explicitly add UTC.
    entries.append(entry_template.substitute(tz_id=0, tz_name="+00:00"))
    with open(args.file_input) as file_in:
        for line in file_in:
            line = line.strip()
            if line[0] == "#":
                continue

            tz_id, tz_name = line.split()
            entries.append(entry_template.substitute(tz_id=tz_id, tz_name=tz_name))
    print(cpp_template.substitute(entries="\n".join(entries)))
    return 0


if __name__ == "__main__":
    sys.exit(main())



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from velox.py.type import BIGINT, ROW
from velox.py.plan_builder import PlanBuilder, PlanNode


class TestPyVeloxPlanBuidler(unittest.TestCase):
    def test_plan_builder(self):
        plan_builder = PlanBuilder()
        self.assertEqual(plan_builder.get_plan_node(), None)

        # Table scan. Needs at least the output schema and it needs to be a ROW.
        self.assertRaises(RuntimeError, plan_builder.table_scan)
        self.assertRaises(RuntimeError, plan_builder.table_scan, BIGINT())

        plan_builder.table_scan(ROW(["c0"], [BIGINT()]))
        table_scan = plan_builder.get_plan_node()
        self.assertTrue(isinstance(table_scan, PlanNode))
        self.assertEqual(table_scan.name(), "TableScan")

        # Project.
        self.assertRaises(
            RuntimeError, plan_builder.project, ["c1"]
        )  # c1 non-existent.
        plan_builder.project(["c0 + 1 AS d0", "100"])
        project = plan_builder.get_plan_node()
        self.assertEqual(project.name(), "Project")

        self.assertNotEqual(table_scan.id(), project.id())

        # Filter.
        self.assertRaises(
            RuntimeError, plan_builder.filter, "c0 > 1"
        )  # c0 non-existent.
        plan_builder.filter("d0 - 10 > 100")
        filter_node = plan_builder.get_plan_node()
        self.assertEqual(filter_node.name(), "Filter")

        self.assertNotEqual(filter_node.id(), project.id())

        self.assertEqual(
            str(filter_node),
            "-- Filter[2]\n" "  -- Project[1]\n" "    -- TableScan[0]\n",
        )

    def test_multiple_plan_builders(self):
        """
        Tests that mutliple plan builders (used to create plans with multiple
        pipelines, like containing joins), generate expected disjoins node ids.
        """
        plan_builder1 = PlanBuilder()
        plan_builder1.table_scan(ROW(["c0"], [BIGINT()]))
        self.assertEqual(plan_builder1.get_plan_node().id(), "0")

        # A new independent plan builder.
        plan_builder2 = PlanBuilder()
        plan_builder2.table_scan(ROW(["c0"], [BIGINT()]))
        self.assertEqual(plan_builder2.get_plan_node().id(), "0")

        # Chained plan builders.
        plan_builder1_1 = plan_builder1.new_builder()
        plan_builder1_1.table_scan(ROW(["c0"], [BIGINT()]))
        self.assertEqual(plan_builder1_1.get_plan_node().id(), "1")

        plan_builder1_2 = plan_builder1.new_builder()
        plan_builder1_2.table_scan(ROW(["c0"], [BIGINT()]))
        self.assertEqual(plan_builder1_2.get_plan_node().id(), "2")

        plan_builder1.project([])
        self.assertEqual(plan_builder1.get_plan_node().id(), "3")



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pyarrow as pa
import pyvelox.pyvelox as pv
import unittest


class TestVeloxVector(unittest.TestCase):
    def test_inheritance(self):
        v1 = pv.from_list([1, 2, 3])
        v2 = pv.from_list(["hello", "world"])
        v3 = pv.constant_vector(1000, 10)
        v4 = pv.dictionary_vector(pv.from_list([1, 2, 3]), [0, 0, 1])

        self.assertTrue(isinstance(v1, pv.BaseVector))
        self.assertTrue(isinstance(v2, pv.BaseVector))
        self.assertTrue(isinstance(v3, pv.BaseVector))
        self.assertTrue(isinstance(v4, pv.BaseVector))

        self.assertTrue(isinstance(v1, pv.SimpleVector_BIGINT))
        self.assertTrue(isinstance(v2, pv.SimpleVector_VARBINARY))
        self.assertTrue(isinstance(v3, pv.SimpleVector_BIGINT))
        self.assertTrue(isinstance(v4, pv.SimpleVector_BIGINT))

        self.assertTrue(isinstance(v1, pv.FlatVector_BIGINT))
        self.assertTrue(isinstance(v2, pv.FlatVector_VARBINARY))
        self.assertTrue(isinstance(v3, pv.ConstantVector_BIGINT))
        self.assertTrue(isinstance(v4, pv.DictionaryVector_BIGINT))

        self.assertFalse(isinstance(v1, pv.ConstantVector_BIGINT))
        self.assertFalse(isinstance(v2, pv.ConstantVector_VARBINARY))
        self.assertFalse(isinstance(v3, pv.FlatVector_BIGINT))
        self.assertFalse(isinstance(v4, pv.ConstantVector_BIGINT))

    def test_from_list(self):
        self.assertTrue(isinstance(pv.from_list([1, 2, 3]), pv.BaseVector))
        self.assertTrue(isinstance(pv.from_list([1, None, None]), pv.BaseVector))
        self.assertTrue(isinstance(pv.from_list(["hello", "world"]), pv.BaseVector))
        with self.assertRaises(TypeError):
            pv.from_list(["hello", 3.14])
        with self.assertRaises(ValueError):
            pv.from_list([None, None, None])
        with self.assertRaises(ValueError):
            pv.from_list([])

    def test_from_list_with_type(self):
        list_a = [0, 1, 3]
        a = pv.from_list(list_a, pv.BooleanType())
        self.assertEqual(a.typeKind().name, "BOOLEAN")
        for i in range(len(a)):
            self.assertTrue(isinstance(a[i], bool))
            self.assertEqual(a[i], bool(list_a[i]))
        self.assertTrue(
            isinstance(
                pv.from_list([None, None, None], pv.VarcharType()), pv.BaseVector
            )
        )
        empty_vector = pv.from_list([], pv.IntegerType())
        self.assertTrue(isinstance(empty_vector, pv.BaseVector))
        with self.assertRaises(IndexError):
            a = empty_vector[0]
        with self.assertRaises(RuntimeError):
            a = pv.from_list(
                [0, 1, 3], pv.VarcharType()
            )  # Conversion not possible from int to varchar
        list_b = [0.2, 1.2, 3.23]
        b = pv.from_list(list_b, pv.RealType())
        for i in range(len(list_b)):
            self.assertNotAlmostEqual(list_b[i], b[i], places=17)

        # dtype as a keyword argument
        integerVector = pv.from_list([1, 3, 11], dtype=pv.IntegerType())
        self.assertTrue(isinstance(integerVector, pv.BaseVector))
        self.assertEqual(integerVector.typeKind().name, "INTEGER")

    def test_constant_encoding(self):
        ints = pv.constant_vector(1000, 10)
        strings = pv.constant_vector("hello", 100)
        null = pv.constant_vector(None, 1000, pv.SmallintType())
        floats = pv.constant_vector(-3.14, 499)
        self.assertEqual(ints.encoding(), pv.VectorEncodingSimple.CONSTANT)
        self.assertEqual(strings.encoding(), pv.VectorEncodingSimple.CONSTANT)
        self.assertEqual(null.encoding(), pv.VectorEncodingSimple.CONSTANT)
        self.assertEqual(floats.encoding(), pv.VectorEncodingSimple.CONSTANT)
        self.assertEqual(len(ints), 10)
        self.assertEqual(len(strings), 100)
        self.assertEqual(len(null), 1000)
        self.assertEqual(len(floats), 499)
        self.assertEqual(ints.typeKind(), pv.TypeKind.BIGINT)
        self.assertEqual(strings.typeKind(), pv.TypeKind.VARCHAR)
        self.assertEqual(null.typeKind(), pv.TypeKind.SMALLINT)
        self.assertEqual(floats.typeKind(), pv.TypeKind.DOUBLE)

        for i in range(len(ints)):
            self.assertEqual(ints[i], 1000)
        for x in ints:
            self.assertEqual(x, 1000)

        for i in range(len(strings)):
            self.assertEqual(strings[i], "hello")
        for x in strings:
            self.assertEqual(x, "hello")

        for i in range(len(null)):
            self.assertEqual(null[i], None)
        for x in null:
            self.assertEqual(x, None)

        for i in range(len(floats)):
            self.assertEqual(floats[i], -3.14)
        for x in floats:
            self.assertEqual(x, -3.14)

        with self.assertRaises(IndexError):
            ints[10]
        with self.assertRaises(TypeError):
            ints[1] = -1

    def test_dictionary_encoding(self):
        base_indices = [0, 0, 1, 0, 2]
        vec = pv.dictionary_vector(pv.from_list([1, 2, 3]), base_indices)
        self.assertTrue(isinstance(vec, pv.DictionaryVector_BIGINT))
        expected_values = [1, 1, 2, 1, 3]
        self.assertEqual(len(vec), len(expected_values))
        for i in range(len(vec)):
            self.assertEqual(vec[i], expected_values[i])

        indices = vec.indices()
        self.assertTrue(isinstance(indices, pv.DictionaryIndices))
        self.assertEqual(len(indices), len(base_indices))
        for i in range(len(indices)):
            self.assertEqual(indices[i], base_indices[i])

        base_vector = pv.from_list([1, 2, 3, 4, 5, 6, 7])
        base_indices = [6, 6]
        vec = pv.dictionary_vector(base_vector, base_indices)
        expected_values = [7, 7]
        for i in range(len(vec)):
            self.assertEqual(vec[i], expected_values[i])

        with self.assertRaises(TypeError):
            pv.dictionary_vector(pv.from_list([1, 2, 3]), ["a", 0, 0, "b"])
        with self.assertRaises(IndexError):
            pv.dictionary_vector(pv.from_list([1, 2, 3]), [1, 2, 1000000])
            pv.dictionary_vector(pv.from_list([1, 2, 3]), [0, -1, -2])

    def test_array_vector(self):
        v1 = pv.from_list([[1, 2, 3], [1, 2, 3]])
        self.assertTrue(isinstance(v1, pv.ArrayVector))
        self.assertTrue(isinstance(v1.elements(), pv.FlatVector_BIGINT))
        self.assertEqual(len(v1), 2)
        expected_flat = [1, 2, 3, 1, 2, 3]
        self.assertEqual(len(expected_flat), len(v1.elements()))
        for i in range(len(expected_flat)):
            self.assertEqual(expected_flat[i], v1.elements()[i])

        v2 = pv.from_list([[1], [1, 2, None]])
        self.assertTrue(isinstance(v2, pv.ArrayVector))
        self.assertTrue(isinstance(v2.elements(), pv.FlatVector_BIGINT))
        self.assertEqual(len(v2), 2)
        expected_flat = [1, 1, 2, None]
        self.assertEqual(len(v2.elements()), len(expected_flat))
        for i in range(len(expected_flat)):
            self.assertEqual(expected_flat[i], v2.elements()[i])

        doubleNested = pv.from_list([[[1, 2], [3, None]], [[1], [2]]])
        self.assertTrue(isinstance(doubleNested, pv.ArrayVector))
        self.assertTrue(isinstance(doubleNested.elements(), pv.ArrayVector))
        self.assertEqual(len(doubleNested), 2)
        elements = doubleNested.elements().elements()
        self.assertTrue(isinstance(elements, pv.FlatVector_BIGINT))
        self.assertEqual(len(elements), 6)
        expected_firstElements = [1, 2, 3, None, 1, 2]
        self.assertEqual(len(elements), len(expected_firstElements))
        for i in range(len(expected_firstElements)):
            self.assertEqual(expected_firstElements[i], elements[i])

        with self.assertRaises(TypeError):
            a = pv.from_list([[[1, 2], [3, 4]], [[1.1], [2.3]]])

        with self.assertRaises(ValueError):
            v = pv.from_list([[None], [None, None, None]])

        with self.assertRaises(TypeError):
            a = pv.from_list([[[1, 2], [3, 4]], [["hello"], ["world"]]])

    def test_to_string(self):
        self.assertEqual(
            str(pv.from_list([1, 2, 3])),
            "[FLAT BIGINT: 3 elements, no nulls]",
        )
        self.assertEqual(
            str(pv.from_list([1, None, 3])),
            "[FLAT BIGINT: 3 elements, 1 nulls]",
        )

    def test_get_item(self):
        ints = pv.from_list([1, 2, None, None, 3])
        self.assertEqual(ints[0], 1)
        self.assertEqual(ints[1], 2)
        self.assertEqual(ints[2], None)
        self.assertEqual(ints[3], None)
        self.assertEqual(ints[4], 3)

        strs = pv.from_list(["hello", "world", None])
        self.assertEqual(strs[0], "hello")
        self.assertEqual(strs[1], "world")
        self.assertEqual(strs[2], None)
        self.assertNotEqual(strs[0], "world")
        self.assertNotEqual(strs[2], "world")

        with self.assertRaises(IndexError):
            ints[5]
        with self.assertRaises(IndexError):
            ints[-1]
        with self.assertRaises(IndexError):
            strs[1000]
        with self.assertRaises(IndexError):
            strs[-1000]

    def test_set_item(self):
        ints = pv.from_list([1, 2, None, None, 3])
        self.assertEqual(ints[2], None)
        ints[2] = 10
        self.assertEqual(ints[2], 10)
        ints[4] = None
        self.assertEqual(ints[4], None)

        strs = pv.from_list(["googly", "doogly"])
        self.assertEqual(strs[1], "doogly")
        strs[1] = "moogly"
        self.assertEqual(strs[1], "moogly")
        strs[0] = None
        self.assertEqual(strs[0], None)

        with self.assertRaises(IndexError):
            ints[5] = 10
        with self.assertRaises(IndexError):
            ints[-1] = 10
        with self.assertRaises(IndexError):
            strs[1000] = "hi"
        with self.assertRaises(IndexError):
            strs[-1000] = "bye"
        with self.assertRaises(TypeError):
            ints[3] = "ni hao"
        with self.assertRaises(TypeError):
            strs[0] = 2

    def test_length(self):
        ints = pv.from_list([1, 2, None])
        self.assertEqual(len(ints), 3)
        self.assertEqual(ints.size(), 3)

        strs = pv.from_list(["hi", "bye"])
        self.assertEqual(len(strs), 2)
        self.assertEqual(strs.size(), 2)

    def test_numeric_limits(self):
        bigger_than_int32 = pv.from_list([1 << 33])
        self.assertEqual(bigger_than_int32[0], 1 << 33)
        with self.assertRaises(RuntimeError):
            bigger_than_int64 = pv.from_list([1 << 63])
        smaller_than_int64 = pv.from_list([(1 << 62) + (1 << 62) - 1])

    def test_type(self):
        ints = pv.from_list([1, 2, None])
        self.assertEqual(ints.dtype(), pv.BigintType())
        self.assertEqual(ints.typeKind(), pv.TypeKind.BIGINT)

        strs = pv.from_list(["a", "b", None])
        self.assertEqual(strs.dtype(), pv.VarcharType())
        self.assertEqual(strs.typeKind(), pv.TypeKind.VARCHAR)

    def test_misc(self):
        ints = pv.from_list([3, 4, 3, None])

        self.assertTrue(ints.mayHaveNulls())

        self.assertFalse(ints.isLazy())

        self.assertTrue(ints.isNullAt(3))
        self.assertFalse(ints.isNullAt(0))
        with self.assertRaises(IndexError):
            ints.isNullAt(10)
        with self.assertRaises(IndexError):
            ints.isNullAt(-10)

        self.assertEqual(ints.hashValueAt(0), ints.hashValueAt(2))
        self.assertNotEqual(ints.hashValueAt(0), ints.hashValueAt(1))
        with self.assertRaises(IndexError):
            ints.hashValueAt(10)
        with self.assertRaises(IndexError):
            ints.hashValueAt(-10)

        self.assertEqual(ints.encoding(), pv.VectorEncodingSimple.FLAT)

    def test_append(self):
        ints1 = pv.from_list([4, 2, 0, None])
        ints2 = pv.from_list([1, 3, 3, 7])
        self.assertEqual(len(ints1), 4)
        ints1.append(ints2)
        self.assertEqual(len(ints1), 8)
        self.assertEqual(ints1[5], 3)
        self.assertEqual(ints1[7], 7)

        strs1 = pv.from_list(["bork", "cork"])
        strs2 = pv.from_list(["pork", "stork"])
        self.assertEqual(len(strs1), 2)
        strs1.append(strs2)
        self.assertEqual(len(strs1), 4)
        self.assertEqual(strs1[3], "stork")

        with self.assertRaises(TypeError):
            ints2.append(strs2)

    def test_slice(self):
        a = pv.from_list(list(range(0, 10)))

        b = a.slice(2, 6)
        self.assertEqual(len(b), 4)
        for i in range(4):
            self.assertEqual(b[i], i + 2)

        with self.assertRaises(NotImplementedError):
            c = a.slice(2, 6, 2)

        d = a[3:6]
        self.assertEqual(len(d), 3)
        for i in range(3):
            self.assertEqual(d[i], i + 3)

        with self.assertRaises(NotImplementedError):
            e = a[3:8:3]

    def test_export_to_arrow(self):
        test_cases = [
            ([1, 2, 3], pa.int64()),
            ([1.1, 2.2, 3.3], pa.float64()),
            (["ab", "bc", "ca"], pa.string()),
        ]
        for data, expected_type in test_cases:
            with self.subTest(data=data):
                vector = pv.from_list(data)
                array = pv.export_to_arrow(vector)

                self.assertEqual(array.type, expected_type)
                self.assertEqual(len(array), len(data))
                self.assertListEqual(array.tolist(), data)

    def test_import_from_arrow(self):
        test_cases = [
            ([11, 26, 31], pa.int64(), pv.IntegerType()),
            ([0.1, 2.5, 3.9], pa.float64(), pv.DoubleType()),
            (["az", "by", "cx"], pa.string(), pv.VarcharType()),
        ]
        for data, dtype, expected_type in test_cases:
            with self.subTest(data=data):
                array = pa.array(data, type=dtype)
                velox_vector = pv.import_from_arrow(array)

                self.assertEqual(velox_vector.size(), len(data))
                self.assertTrue(velox_vector.dtype(), expected_type)
                for i in range(0, len(data)):
                    self.assertEqual(velox_vector[i], data[i])

    def test_roundtrip_conversion(self):
        test_cases = [
            ([41, 92, 13], pv.IntegerType()),
            ([17.19, 22.25, 13.3], pv.DoubleType()),
            (["aa1", "bb2", "cc3"], pv.VarcharType()),
        ]
        for data, expected_type in test_cases:
            with self.subTest(data=data):
                vector = pv.from_list(data)
                array = pv.export_to_arrow(vector)

                velox_vector = pv.import_from_arrow(array)
                self.assertEqual(velox_vector.size(), len(data))
                self.assertTrue(velox_vector.dtype(), expected_type)
                for i in range(0, len(data)):
                    self.assertEqual(velox_vector[i], data[i])

    def test_row_vector_basic(self):
        vals = [
            pv.from_list([1, 2, 3]),
            pv.from_list([4.0, 5.0, 6.0]),
            pv.from_list(["a", "b", "c"]),
        ]

        col_names = ["x", "y", "z"]
        rw = pv.row_vector(col_names, vals)
        rw_str = str(rw)
        expected_str = "0: {1, 4, a}\n1: {2, 5, b}\n2: {3, 6, c}"
        assert expected_str == rw_str

    def test_row_vector_with_nulls(self):
        vals = [
            pv.from_list([1, 2, 3, 1, 2]),
            pv.from_list([4, 5, 6, 4, 5]),
            pv.from_list([7, 8, 9, 7, 8]),
            pv.from_list([10, 11, 12, 10, 11]),
        ]

        col_names = ["a", "b", "c", "d"]
        rw = pv.row_vector(col_names, vals, {0: True, 2: True})
        rw_str = str(rw)
        expected_str = (
            "0: null\n1: {2, 5, 8, 11}\n2: null\n3: {1, 4, 7, 10}\n4: {2, 5, 8, 11}"
        )
        assert expected_str == rw_str

    def test_row_vector_comparison(self):
        u = [
            pv.from_list([1, 2, 3]),
            pv.from_list([7, 4, 9]),
            pv.from_list([10, 11, 12]),
        ]

        v = [
            pv.from_list([1, 2, 3]),
            pv.from_list([7, 8, 9]),
            pv.from_list([10, 11, 12]),
        ]

        w = [
            pv.from_list([1, 2, 3]),
            pv.from_list([7, 8, 9]),
        ]

        u_names = ["a", "b", "c"]
        w_names = ["x", "y"]
        u_rw = pv.row_vector(u_names, u)
        v_rw = pv.row_vector(u_names, v)
        w_rw = pv.row_vector(w_names, w)
        y_rw = pv.row_vector(u_names, u)
        x1_rw = pv.row_vector(u_names, u, {0: True, 2: True})
        x2_rw = pv.row_vector(u_names, u, {0: True, 2: True})

        assert u_rw != w_rw  # num of children doesn't match
        assert u_rw != v_rw  # data doesn't match
        assert u_rw == y_rw  # data match
        assert x1_rw == x2_rw  # with null
        assert x1_rw != u_rw  # with and without null



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import json
import os
import re
import sys
from typing import Any

from deepdiff import DeepDiff

import pyvelox.pyvelox as pv

# Utility to export and diff function signatures.


# From https://stackoverflow.com/questions/287871/how-do-i-print-colored-text-to-the-terminal
class bcolors:
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    BOLD = "\033[1m"


aggregate_pattern = re.compile("(.*)(_merge|_merge_extract|_partial)")


def get_error_string(error_message):
    return f"""
Incompatible changes in function signatures have been detected.

{error_message}

Changing or removing function signatures breaks backwards compatibility as some users may rely on function signatures that no longer exist.

"""


def set_gh_output(name: str, value: Any):
    """Sets a Github Actions output variable. Only single line values are supported.
    value will be converted to a lower case string."""
    value = str(value).lower()

    if "\n" in value:
        raise ValueError("Only single line values are supported.")

    with open(os.environ["GITHUB_OUTPUT"], "a") as f:
        f.write(f"{name}={value}\n")


def show_error(error_message, error_path):
    if error_path:
        with open(error_path, "a+") as f:
            f.writelines(get_error_string(error_message))

    print(get_error_string(error_message))


def export(args):
    """Exports Velox function signatures."""
    pv.clear_signatures()

    if args.spark:
        pv.register_spark_signatures()

    if args.presto:
        pv.register_presto_signatures()

    signatures = pv.get_function_signatures()

    # Convert signatures to json
    jsoned_signatures = {}
    for key in signatures.keys():
        jsoned_signatures[key] = [str(value) for value in signatures[key]]

    # Persist to file
    with open(args.output_file, "w") as f:
        json.dump(jsoned_signatures, f)

    return 0


def export_aggregates(args):
    """Exports Velox Aggregate function signatures."""
    pv.clear_aggregate_signatures()

    if args.spark:
        pv.register_spark_aggregate_signatures()

    if args.presto:
        pv.register_presto_aggregate_signatures()

    signatures = pv.get_aggregate_function_signatures()

    # Convert signatures to json
    jsoned_signatures = {}
    for key in signatures.keys():
        jsoned_signatures[key] = [str(value) for value in signatures[key]]

    # Persist to file
    json.dump(jsoned_signatures, args.output_file)
    return 0


def diff_signatures(base_signatures, contender_signatures, error_path=""):
    """Diffs Velox function signatures. Returns a tuple of the delta diff and exit status"""

    delta = DeepDiff(
        base_signatures,
        contender_signatures,
        ignore_order=True,
        cutoff_distance_for_pairs=0.9,
        report_repetition=True,
        view="tree",
    )
    exit_status = 0
    if delta:
        if "dictionary_item_removed" in delta:
            error_message = ""
            for dic_removed in delta["dictionary_item_removed"]:
                error_message += (
                    f"""Function '{dic_removed.get_root_key()}' has been removed.\n"""
                )
            show_error(error_message, error_path)
            exit_status = 1

        if "values_changed" in delta:
            error_message = ""
            for value_change in delta["values_changed"]:
                error_message += f"""'{value_change.get_root_key()}{value_change.t1}' is changed to '{value_change.get_root_key()}{value_change.t2}'.\n"""
            show_error(error_message, error_path)
            exit_status = 1

        if "repetition_change" in delta:
            error_message = ""
            for rep_change in delta["repetition_change"]:
                error_message += f"""'{rep_change.get_root_key()}{rep_change.t1}' is repeated {rep_change.repetition['new_repeat']} times.\n"""
            show_error(error_message, error_path)
            exit_status = 1

        if "iterable_item_removed" in delta:
            error_message = ""
            for iter_change in delta["iterable_item_removed"]:
                error_message += f"""{iter_change.get_root_key()} has its function signature '{iter_change.t1}' removed.\n"""
            show_error(error_message, error_path)
            exit_status = 1

    else:
        print(f"{bcolors.BOLD}No differences found.")

    return delta, exit_status


def diff(args):
    """Diffs Velox function signatures."""
    with open(args.base) as f:
        base_signatures = json.load(f)

    with open(args.contender) as f:
        contender_signatures = json.load(f)
    return diff_signatures(base_signatures, contender_signatures)[1]


def bias(args):
    with open(args.base) as f:
        base_signatures = json.load(f)

    with open(args.contender) as f:
        contender_signatures = json.load(f)

    tickets = args.ticket_value
    bias_output, status = bias_signatures(
        base_signatures, contender_signatures, tickets, args.error_path
    )

    if bias_output:
        with open(args.output_path, "w") as f:
            print(f"{bias_output}", file=f, end="")

    return status


def bias_signatures(base_signatures, contender_signatures, tickets, error_path):
    """Returns newly added functions as string and a status flag.
    Newly added functions are biased like so `fn_name1=<ticket_count>,fn_name2=<ticket_count>`.
    If it detects incompatible changes returns 1 in the status.
    """
    delta, status = diff_signatures(base_signatures, contender_signatures, error_path)

    if not delta:
        print(f"{bcolors.BOLD} No changes detected: Nothing to do!")
        return "", status

    function_set = set()
    for items in delta.values():
        for item in items:
            function_set.add(item.get_root_key())

    if function_set:
        return f"{f'={tickets},'.join(sorted(function_set)) + f'={tickets}'}", status

    return "", status


def bias_aggregates(args):
    """
    Finds and exports aggregates whose signatures have been modified agasint a baseline.
    Saves the results to a file and sets a Github Actions Output.
    Currently this is hardcoded to presto aggregates.
    """
    with open(args.base) as f:
        base_signatures = json.load(f)

    with open(args.contender) as f:
        contender_signatures = json.load(f)

    delta, status = diff_signatures(
        base_signatures, contender_signatures, args.error_path
    )

    set_gh_output("presto_aggregate_error", status == 1)

    if not delta:
        print(f"{bcolors.BOLD} No changes detected: Nothing to do!")
        return status

    function_set = set()
    for items in delta.values():
        for item in items:
            fn_name = item.get_root_key()
            pattern = aggregate_pattern.match(fn_name)
            if pattern:
                function_set.add(pattern.group(1))
            else:
                function_set.add(fn_name)

    if function_set:
        biased_functions = ",".join(function_set)
        with open(args.output_path, "w") as f:
            print(f"{biased_functions}", file=f, end="")

        set_gh_output("presto_aggregate_functions", True)

    return 0


def gh_bias_check(args):
    """
    Exports signatures for the given group(s) and checks them for changes compared to a baseline.
    Saves the results to a file and sets a Github Actions Output for each group.
    """
    if not os.getenv("GITHUB_ACTIONS"):
        print("This command is meant to be run in a Github Actions environment.")
        return 1

    # export signatures for each group
    for group in args.group:
        print(f"Exporting {group} signatures...")
        export_args = parse_args(
            [
                "export",
                f"--{group}",
                os.path.join(args.signature_dir, group + args.contender_postfix),
            ]
        )
        export(export_args)

    # compare signatures for each group
    for group in args.group:
        print(f"Comparing {group} signatures...")
        bias_args = parse_args(
            [
                "bias",
                os.path.join(args.signature_dir, group + args.base_postfix),
                os.path.join(args.signature_dir, group + args.contender_postfix),
                os.path.join(args.signature_dir, group + args.output_postfix),
                os.path.join(args.signature_dir, group + "_errors"),
            ]
        )

        bias_status = bias(bias_args)
        set_gh_output(f"{group}_error", bias_status == 1)

        # check if there are any changes that require the bias fuzzer to run
        has_tickets = os.path.isfile(
            os.path.join(args.signature_dir, group + args.output_postfix)
        )
        set_gh_output(f"{group}_functions", has_tickets)


def get_tickets(val):
    tickets = int(val)
    if tickets < 0:
        raise argparse.ArgumentTypeError("Cant have negative values!")
    return tickets


def parse_args(args):
    global parser

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description="""Velox Function Signature Utility""",
    )

    command = parser.add_subparsers(dest="command")
    export_command_parser = command.add_parser("export")
    export_command_parser.add_argument("--spark", action="store_true")
    export_command_parser.add_argument("--presto", action="store_true")
    export_command_parser.add_argument("output_file", type=str)

    export_aggregates_command_parser = command.add_parser("export_aggregates")
    export_aggregates_command_parser.add_argument("--spark", action="store_true")
    export_aggregates_command_parser.add_argument("--presto", action="store_true")
    export_aggregates_command_parser.add_argument(
        "output_file", type=argparse.FileType("w")
    )

    diff_command_parser = command.add_parser("diff")
    diff_command_parser.add_argument("base", type=str)
    diff_command_parser.add_argument("contender", type=str)

    bias_command_parser = command.add_parser("bias")
    bias_command_parser.add_argument("base", type=str)
    bias_command_parser.add_argument("contender", type=str)
    bias_command_parser.add_argument("output_path", type=str)
    bias_command_parser.add_argument(
        "ticket_value", type=get_tickets, default=10, nargs="?"
    )
    bias_command_parser.add_argument("error_path", type=str, default="")

    gh_command_parser = command.add_parser("gh_bias_check")
    gh_command_parser.add_argument(
        "group",
        nargs="+",
        help='One or more group names to check for changed signatures. e.g. "spark" or "presto"',
        type=str,
    )
    gh_command_parser.add_argument(
        "--signature_dir", type=str, default="/tmp/signatures"
    )
    gh_command_parser.add_argument(
        "--base_postfix", type=str, default="_signatures_main.json"
    )
    gh_command_parser.add_argument(
        "--contender_postfix", type=str, default="_signatures_contender.json"
    )
    gh_command_parser.add_argument(
        "--output_postfix", type=str, default="_bias_functions"
    )

    bias_aggregate_command_parser = command.add_parser("bias_aggregates")
    bias_aggregate_command_parser.add_argument("base", type=str)
    bias_aggregate_command_parser.add_argument("contender", type=str)
    bias_aggregate_command_parser.add_argument("output_path", type=str)
    bias_aggregate_command_parser.add_argument("error_path", type=str, default="")

    parser.set_defaults(command="help")

    return parser.parse_args(args)


def main():
    args = parse_args(sys.argv[1:])
    return globals()[args.command](args)


if __name__ == "__main__":
    sys.exit(main())



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from velox.py.file import PARQUET, DWRF, NIMBLE, ORC, JSON, TEXT


class TestPyVeloxFile(unittest.TestCase):
    def test_input_file(self):
        file_parquet = PARQUET("my_parquet_file")
        file_nimble = NIMBLE("/tmp/my_nimble")
        file_dwrf = DWRF("dir/some_file.dwrf")
        file_orc = ORC("orc_file")
        file_json = JSON("dir2/other_file.json")
        file_text = TEXT("a/b/csome_file.txt")

        self.assertEqual(str(file_parquet), "my_parquet_file (parquet)")
        self.assertEqual(str(file_nimble), "/tmp/my_nimble (nimble)")
        self.assertEqual(str(file_dwrf), "dir/some_file.dwrf (dwrf)")
        self.assertEqual(str(file_orc), "orc_file (orc)")
        self.assertEqual(str(file_json), "dir2/other_file.json (json)")
        self.assertEqual(str(file_text), "a/b/csome_file.txt (text)")



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import unittest.mock
from scripts.signature import bias_signatures, get_error_string
from pathlib import Path
import json
import io


def read_from_file(file_path):
    return Path(file_path).read_text()


def test_bias(base_signatures, contender_signatures):
    return bias_signatures(
        json.loads(base_signatures), json.loads(contender_signatures), 10
    )


class SignatureTest(unittest.TestCase):
    def test_bias(self):
        # Remove a signature
        _, return_value = test_bias(
            """{"reverse": ["(array(T)) -> array(T)"]}""",
            """{"reverse": []}""",
        )

        self.assertEqual(return_value, 1)

        # Add a new signature
        bias_functions, _ = test_bias(
            """{"reverse": ["(array(T)) -> array(T)"]}""",
            """{"reverse": ["(array(T)) -> array(T)"],
                   "foo": ["(varchar) -> varchar"]}""",
        )

        self.assertEqual(bias_functions, "foo=10")

        # Modify a signature.
        bias_functions, _ = test_bias(
            """{"reverse": ["(array(T)) -> array(T)"]}""",
            """{"reverse": ["(array(T)) -> array(T)", "(varchar) -> varchar"]}""",
        )

        self.assertEqual(bias_functions, "reverse=10")

        # Add more than one signature change
        bias_functions, _ = test_bias(
            """{"reverse": ["(array(T)) -> array(T)"]}""",
            """{"reverse": ["(array(T)) -> array(T)"],
                   "foo": ["(varchar) -> varchar"],
                   "bar": ["(varchar) -> varchar"]}""",
        )

        self.assertEqual(bias_functions, "bar=10,foo=10")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def get_bias_messaging(self, base_signatures, contender_signatures, mock_stdout):
        test_bias(base_signatures, contender_signatures)
        return mock_stdout.getvalue()

    def assert_messaging(self, base_signatures, contender_signatures, expected_message):
        test_bias(base_signatures, contender_signatures)
        actual = self.get_bias_messaging(base_signatures, contender_signatures)
        expected = get_error_string(expected_message)
        expected += "\n"  # Add trailing newline for std output.
        self.assertEquals(expected, actual)

    def test_messaging(self):
        # Remove a signature
        self.assert_messaging(
            """{"reverse": ["(array(T)) -> array(T)"]}""",
            """{"reverse": []}""",
            "reverse has its function signature '(array(T)) -> array(T)' removed.\n",
        )

        # Remove more than one signature
        self.assert_messaging(
            """{"reverse": ["(array(T)) -> array(T)", "(varchar) -> varchar"]}""",
            """{"reverse": []}""",
            """reverse has its function signature '(array(T)) -> array(T)' removed.\nreverse has its function signature '(varchar) -> varchar' removed.\n""",
        )

        # Mutate a signature
        self.assert_messaging(
            """{"reverse": ["(array(T)) -> array(T)"]}""",
            """{"reverse": ["(array(T)) -> array(varchar)"]}""",
            """'reverse(array(T)) -> array(T)' is changed to 'reverse(array(T)) -> array(varchar)'.\n""",
        )

        # Function repeated
        self.assert_messaging(
            """{"reverse": ["(array(T)) -> array(T)"]}""",
            """{"reverse": ["(array(T)) -> array(T)", "(array(T)) -> array(T)"]}""",
            "'reverse(array(T)) -> array(T)' is repeated 2 times.\n",
        )

        # Remove a udf
        self.assert_messaging(
            """{"reverse": ["(array(T)) -> array(T)"]}""",
            """{}""",
            "Function 'reverse' has been removed.\n",
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# @nolint
# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2024 by Rivos Inc.
# Licensed under the Apache License, Version 2.0, see LICENSE for details.
# SPDX-License-Identifier: Apache-2.0

from generator_common import (
    args,
    clang_format,
    extract_node,
    parse,
    read_extent,
    to_snake_case,
    CursorKind,
    COPYRIGHT,
)

import io
import subprocess
from abc import ABC, abstractmethod


AUTOGEN_HEADER = f"""/*
 * This file is auto-generated from kernel_generator.py
 * DO NOT EDIT!
 */
"""


def kernels_extractor(tu):
    kernels = []
    for c in tu.cursor.get_children():
        if c.displayname == "kernels" and c.kind == CursorKind.NAMESPACE:
            kernels = list(map(extract_node, c.get_children()))
    return kernels


def includes_extractor(tu):
    return list(
        filter(
            lambda x: x.kind == CursorKind.INCLUSION_DIRECTIVE, tu.cursor.get_children()
        )
    )


class KernelCodeGen(ABC):
    @property
    @abstractmethod
    def includes(self): ...

    @property
    @abstractmethod
    def platform_type(self): ...

    def num_warp_threads(self):
        return None

    @property
    def use_namespace(self):
        return True

    def get_templates_postfix(self, attrs):
        return []

    def write_func_attrs(self, out, attrs):
        pass

    def get_func_name(self, name):
        return name

    def get_args_prefix(self, attrs):
        return []

    def get_args(self, fn_props):
        args = list(map(lambda x: read_extent(x.extent), fn_props["function_params"]))
        return args

    def write_func_prologue(self, out, attrs):
        pass

    def generate(self, tu, filename):
        out = io.StringIO()
        # Copyright, autogen header, and platform specific includes
        out.write(f"{COPYRIGHT}\n{AUTOGEN_HEADER}\n{self.includes}\n")
        # Add general includes
        for i in includes_extractor(tu):
            out.write(read_extent(i.extent))
            out.write("\n")
        out.write("\n")
        # parse out the kernels and rewrite them in platform's style
        kernels = kernels_extractor(tu)
        if self.use_namespace:
            out.write("namespace kernels {\n\n")
        warp_threads = self.num_warp_threads()
        if warp_threads != None:
            out.write(f"enum {{ WARP_THREADS = {warp_threads} }};")
        for kernel in kernels:
            kernel_name = kernel["spelling"]
            fn_props = kernel["fn_props"]
            template_params = fn_props["template_params"]
            attrs = fn_props["attrs"]
            body = fn_props["body"]

            # Consistency checks
            assert (kernel["kind"] == CursorKind.FUNCTION_TEMPLATE) == (
                len(template_params) != 0
            )

            # Add template parameters if applicable
            extra_template_args = self.get_templates_postfix(attrs)
            attrs["has_template_params"] = False
            if len(template_params) != 0 or len(extra_template_args) != 0:
                attrs["has_template_params"] = True
                out.write("template <")
                out.write(
                    ", ".join(
                        list(map(lambda x: read_extent(x.extent), template_params))
                        + extra_template_args
                    )
                )
                out.write(">\n")

            # Emit any function attributes needed
            self.write_func_attrs(out, attrs)

            # Write out function return type, name and arguments
            out.write(fn_props["return_type"].spelling)
            out.write(" ")
            out.write(self.get_func_name(kernel_name))
            out.write("(")
            out.write(", ".join(self.get_args_prefix(attrs) + self.get_args(fn_props)))
            out.write(")\n")

            # Begin the body
            out.write("{\n")
            self.write_func_prologue(out, attrs)
            out.write(read_extent(body.extent)[1:-1])
            out.write("}\n\n")
        if self.use_namespace:
            out.write("}  // namespace kernels\n")

        # Write out to specified file and clang-format the result
        with open(filename, "w") as outf:
            outf.write(out.getvalue())
        subprocess.run([clang_format, "-i", filename])


class OpenclBackend(KernelCodeGen):
    @property
    def includes(self):
        return '#include "breeze/platforms/opencl.h"\n'

    @property
    def platform_type(self):
        return "OpenCLPlatform<BLOCK_THREADS, WARP_THREADS>"

    @property
    def use_namespace(self):
        return False

    def num_warp_threads(self):
        return 32

    def write_func_attrs(self, out, attrs):
        # If this function isn't templated we may emit it as a kernel directly.
        if not attrs.get("has_template_params"):
            out.write("kernel ")

    def get_func_name(self, name):
        return to_snake_case(name)

    def get_templates_postfix(self, attrs):
        shared_mems = attrs.get("SharedMem", [])
        assert len(shared_mems) <= 1
        if shared_mems:
            return ["typename SharedMemType"]
        else:
            return []

    def write_func_prologue(self, out, attrs):
        platform_names = attrs.get("PlatformName", [])
        assert len(platform_names) <= 1
        for p in platform_names:
            out.write(f"  using PlatformT = {self.platform_type};")
            out.write(f"  PlatformT {p};")

    def get_args(self, fn_props):
        args = super().get_args(fn_props)

        # All pointers should be made annotated global
        args = list(map(lambda x: "global " + x if "*" in x else x, args))

        # shared_mem always precedes num_items (if it exists).
        shared_mems = fn_props["attrs"].get("SharedMem", [])
        assert len(shared_mems) <= 1
        if shared_mems:
            (_type, ident) = shared_mems[0].split(";", 1)
            shmem_arg = f"local SharedMemType* {ident}"
            if args[-1] == "int num_items":
                args = args[:-1] + [shmem_arg, args[-1]]
            else:
                args.append(shmem_arg)

        return args


class MetalBackend(KernelCodeGen):
    @property
    def includes(self):
        return '#include <metal_stdlib>\n#include "breeze/platforms/metal.h"\n'

    @property
    def platform_type(self):
        return "MetalPlatform<BLOCK_THREADS, WARP_THREADS>"

    @property
    def use_namespace(self):
        return False

    def num_warp_threads(self):
        return 32

    def write_func_attrs(self, out, attrs):
        # If this function isn't templated we may emit it as a kernel directly.
        if not attrs.get("has_template_params"):
            out.write("kernel ")

    def get_func_name(self, name):
        return to_snake_case(name)

    def get_templates_postfix(self, attrs):
        platform_names = attrs.get("PlatformName", [])
        shared_mems = attrs.get("SharedMem", [])
        assert len(platform_names) <= 1
        assert len(shared_mems) <= 1
        postfix = []
        if platform_names:
            postfix += ["typename PlatformT"]
        if shared_mems:
            postfix += ["typename SharedMemType"]
        return postfix

    def get_args_prefix(self, attrs):
        platform_names = attrs.get("PlatformName", [])
        assert len(platform_names) <= 1
        prefix = []
        if platform_names:
            prefix += [f"PlatformT {platform_names[0]}"]
        return prefix

    def get_args(self, fn_props):
        args = super().get_args(fn_props)

        # All pointers should be made annotated device
        args = list(map(lambda x: "device " + x if "*" in x else x, args))

        # shared_mem always precedes num_items (if it exists).
        shared_mems = fn_props["attrs"].get("SharedMem", [])
        assert len(shared_mems) <= 1
        if shared_mems:
            (_type, ident) = shared_mems[0].split(";", 1)
            shmem_arg = f"threadgroup SharedMemType* {ident}"
            if args[-1] == "int num_items":
                args = args[:-1] + [shmem_arg, args[-1]]
            else:
                args.append(shmem_arg)

        return args


class OpenmpBackend(KernelCodeGen):
    @property
    def includes(self):
        return '#include "breeze/platforms/openmp.h"\n'

    @property
    def platform_type(self):
        return "OpenMPPlatform<BLOCK_THREADS, BLOCK_THREADS>"

    def get_templates_postfix(self, attrs):
        platform_names = attrs.get("PlatformName", [])
        shared_mems = attrs.get("SharedMem", [])
        assert len(platform_names) <= 1
        assert len(shared_mems) <= 1
        postfix = []
        if shared_mems:
            postfix += ["typename SharedMemType"]
        if platform_names:
            postfix += [f"typename PlatformT = {self.platform_type}"]
        return postfix

    def get_args_prefix(self, attrs):
        platform_names = attrs.get("PlatformName", [])
        shared_mems = attrs.get("SharedMem", [])
        assert len(platform_names) <= 1
        assert len(shared_mems) <= 1
        prefix = []
        if platform_names:
            prefix += [f"PlatformT {platform_names[0]}"]
        if shared_mems:
            (_type, ident) = shared_mems[0].split(";", 1)
            prefix += [f"SharedMemType* {ident}"]
        return prefix


class SyclBackend(KernelCodeGen):
    @property
    def includes(self):
        return '#include "breeze/platforms/sycl.hpp"\n'

    @property
    def platform_type(self):
        return "SyCLPlatform<BLOCK_THREADS, WARP_THREADS>"

    def num_warp_threads(self):
        return 32

    def get_templates_postfix(self, attrs):
        shared_mems = attrs.get("SharedMem", [])
        assert len(shared_mems) <= 1
        if shared_mems:
            return ["typename SharedMemType"]
        else:
            return []

    def get_args_prefix(self, attrs):
        platform_names = attrs.get("PlatformName", [])
        assert len(platform_names) <= 1
        return ["cl::sycl::nd_item<1> item"]

    def get_args(self, fn_props):
        args = super().get_args(fn_props)

        # shared_mem always precedes num_items (if it exists).
        shared_mems = fn_props["attrs"].get("SharedMem", [])
        assert len(shared_mems) <= 1
        if shared_mems:
            (_type, ident) = shared_mems[0].split(";", 1)
            shmem_arg = f"SharedMemType* {ident}"
            if args[-1] == "int num_items":
                args = args[:-1] + [shmem_arg, args[-1]]
            else:
                args.append(shmem_arg)

        return args

    def write_func_prologue(self, out, attrs):
        platform_names = attrs.get("PlatformName", [])
        assert len(platform_names) <= 1
        for p in platform_names:
            out.write(f"  using PlatformT = {self.platform_type};")
            out.write(f"  PlatformT {p}{{item}};")


class CudaBackend(KernelCodeGen):
    @property
    def includes(self):
        return '#include "breeze/platforms/cuda.cuh"\n'

    @property
    def platform_type(self):
        return "CudaPlatform<BLOCK_THREADS, WARP_THREADS>"

    def num_warp_threads(self):
        return 32

    def write_func_attrs(self, out, attrs):
        # Add __global__ since this is a kernel
        out.write("__global__ ")

    def write_func_prologue(self, out, attrs):
        platform_names = attrs.get("PlatformName", [])
        assert len(platform_names) <= 1
        for p in platform_names:
            out.write(f"  using PlatformT = {self.platform_type};")
            out.write(f"  PlatformT {p};")
        shared_mems = attrs.get("SharedMem", [])
        for shared_mem in shared_mems:
            (type, ident) = shared_mem.split(";", 1)
            out.write(f"  __shared__ {type} {ident}_;\n")
            out.write(f"  auto {ident} = ({type}*)&{ident}_;\n")


class HipBackend(CudaBackend):
    @property
    def includes(self):
        return '#include "breeze/platforms/hip.hpp"\n'

    @property
    def platform_type(self):
        return "HipPlatform<BLOCK_THREADS, WARP_THREADS>"

    def num_warp_threads(self):
        return 32

    def write_func_attrs(self, out, attrs):
        # Add __global__ since this is a kernel
        out.write("__global__ ")


if __name__ == "__main__":
    index, tu = parse(args.template)
    # Normally it would be good to check tu.diagnostics here to ensure
    # there are no errors or other diagnostic messages outstanding,
    # thereby assuring that all the input is being parsed as expected.
    # However by nature of the function bodies being incomplete
    # (due to Platform and Shared symbols being factored out-of-line)
    # there will always be some error reported and so we disregard the
    # diagnostics and proceed optimistically.
    if args.backend == "cuda":
        CudaBackend().generate(tu, args.out)
    elif args.backend == "hip":
        HipBackend().generate(tu, args.out)
    elif args.backend == "sycl":
        SyclBackend().generate(tu, args.out)
    elif args.backend == "openmp":
        OpenmpBackend().generate(tu, args.out)
    elif args.backend == "opencl":
        OpenclBackend().generate(tu, args.out)
    elif args.backend == "metal":
        MetalBackend().generate(tu, args.out)
    else:
        raise ValueError(f"Unknown backend provided: {args.backend}")



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
from collections import OrderedDict
import os
import regex
import subprocess
import sys

from util import attrdict
import util

EXTENSIONS = "cpp,h,inc,prolog"
SCRIPTS = util.script_path()


def get_diff(file, formatted):
    if not formatted.endswith("\n"):
        formatted = formatted + "\n"

    status, stdout, stderr = util.run(
        f"diff -u {file} --label {file} --label {file} -", input=formatted
    )
    if stdout != "":
        stdout = f"diff a/{file} b/{file}\n" + stdout

    return status, stdout, stderr


class CppFormatter(str):
    def diff(self, commit):
        if commit == "":
            return get_diff(self, util.run(f"clang-format --style=file {self}")[1])
        else:
            return util.run(
                f"{SCRIPTS}/git-clang-format -q --extensions='{EXTENSIONS}' --diff --style=file {commit} {self}"
            )

    def fix(self, commit):
        if commit == "":
            return util.run(f"clang-format -i --style=file {self}")[0] == 0
        else:
            return (
                util.run(
                    f"{SCRIPTS}/git-clang-format -q --extensions='{EXTENSIONS}' --style=file {commit} {self}"
                )[0]
                == 0
            )


class CMakeFormatter(str):
    def __init__(self, commit) -> None:
        super().__init__()
        try:
            import yaml
        except ModuleNotFoundError:
            # We need pyyaml so cmake-format can read '.cmake-format.yml'
            # otherwise it will run with default
            raise SystemExit("Please install 'pyyaml' for the CMake formatter.")

    def diff(self, commit):
        return get_diff(
            self, util.run(f"cmake-format --first-comment-is-literal True {self}")[1]
        )

    def fix(self, commit):
        return (
            util.run(f"cmake-format --first-comment-is-literal True -i {self}")[0] == 0
        )


class PythonFormatter(str):
    def diff(self, commit):
        return util.run(f"black -q --diff {self}")

    def fix(self, commit):
        return util.run(f"black -q {self}")[0] == 0


format_file_types = OrderedDict(
    {
        "CMakeLists.txt": attrdict({"formatter": CMakeFormatter}),
        "*.cmake": attrdict({"formatter": CMakeFormatter}),
        "*.cpp": attrdict({"formatter": CppFormatter}),
        "*.h": attrdict({"formatter": CppFormatter}),
        "*.inc": attrdict({"formatter": CppFormatter}),
        "*.prolog": attrdict({"formatter": CppFormatter}),
        "*.hpp": attrdict({"formatter": CppFormatter}),
        "*.cu": attrdict({"formatter": CppFormatter}),
        "*.cuh": attrdict({"formatter": CppFormatter}),
        "*.clcpp": attrdict({"formatter": CppFormatter}),
        "*.mm": attrdict({"formatter": CppFormatter}),
        "*.metal": attrdict({"formatter": CppFormatter}),
        "*.py": attrdict({"formatter": PythonFormatter}),
    }
)


def get_formatter(filename):
    if filename in format_file_types:
        return format_file_types[filename]

    return format_file_types.get("*" + util.get_fileextn(filename), None)


def format_command(commit, files, fix):
    ok = 0
    for filepath in files:
        filename = util.get_filename(filepath)
        filetype = get_formatter(filename)

        if filetype is None:
            print("Skip : " + filepath, file=sys.stderr)
            continue

        file = filetype.formatter(filepath)

        if fix == "show":
            status, diff, stderr = file.diff(commit)

            if stderr != "":
                ok = 1
                print(f"Error: {file}", file=sys.stderr)
                continue

            if diff != "" and diff != "no modified files to format":
                ok = 1
                print(f"Fix  : {file}", file=sys.stderr)
                print(diff)
            else:
                print(f"Ok   : {file}", file=sys.stderr)

        else:
            print(f"Fix  : {file}", file=sys.stderr)
            if not file.fix(commit):
                ok = 1
                print(f"Error: {file}", file=sys.stderr)

    return ok


def header_command(commit, files, fix):
    options = "-vk" if fix == "show" else "-i"

    status, stdout, stderr = util.run(
        f"{SCRIPTS}/license-header.py {options} -", input=files
    )

    if stdout != "":
        print(stdout)

    return status


def tidy_command(commit, files, fix):
    files = [file for file in files if regex.match(r".*\.cpp$", file)]

    if not files:
        return 0

    commit = f"--commit {commit}" if commit != "" else ""
    fix = "--fix" if fix == "fix" else ""

    status, stdout, stderr = util.run(
        f"{SCRIPTS}/run-clang-tidy.py {commit} {fix} -", input=files
    )

    if stdout != "":
        print(stdout)

    return status


def get_commit(files):
    if files == "commit":
        return "HEAD^"

    if files == "main" or files == "master":
        return util.run(f"git merge-base origin/{files} HEAD")[1]

    return ""


def get_files(commit, path):
    filelist = []

    if commit != "":
        status, stdout, stderr = util.run(
            f"git diff --relative --name-only --diff-filter='ACMR' {commit}"
        )
        filelist = stdout.splitlines()
    else:
        if os.path.isfile(path):
            filelist.append(path)
        else:
            for root, dirs, files in os.walk(path):
                for name in files:
                    filelist.append(os.path.join(root, name))

    return [
        file
        for file in filelist
        if "/data/" not in file
        and "velox/external/" not in file
        and "build/fbcode_builder" not in file
        and "build/deps" not in file
        and "cmake-build-debug" not in file
        and "NOTICE.txt" != file
        and "velox/docs/affiliations_map.txt" != file
    ]


def help(args):
    parser.print_help()
    return 0


def add_check_options(subparser, name):
    parser = subparser.add_parser(name)
    parser.add_argument("--fix", action="store_const", default="show", const="fix")
    return parser


def add_options(parser):
    files = parser.add_subparsers(dest="files")

    tree_parser = add_check_options(files, "tree")
    tree_parser.add_argument("path", default="")

    branch_parser = add_check_options(files, "main")
    branch_parser = add_check_options(files, "master")
    commit_parser = add_check_options(files, "commit")


def add_check_command(parser, name):
    subparser = parser.add_parser(name)
    add_options(subparser)

    return subparser


def parse_args():
    global parser
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description="""Check format/header/tidy

    check.py {format,header,tidy} {commit,branch} [--fix]
    check.py {format,header,tidy} {tree} [--fix] PATH
""",
    )
    command = parser.add_subparsers(dest="command")
    command.add_parser("help")

    format_command_parser = add_check_command(command, "format")
    header_command_parser = add_check_command(command, "header")
    tidy_command_parser = add_check_command(command, "tidy")

    parser.set_defaults(path="")
    parser.set_defaults(command="help")

    return parser.parse_args()


def run_command(args, command):
    commit = get_commit(args.files)
    files = get_files(commit, args.path)

    return command(commit, files, args.fix)


def format(args):
    return run_command(args, format_command)


def header(args):
    return run_command(args, header_command)


def tidy(args):
    return run_command(args, tidy_command)


def main():
    args = parse_args()
    return globals()[args.command](args)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from collections import OrderedDict
import fnmatch
import os
import regex
import sys


class attrdict(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


def parse_args():
    parser = argparse.ArgumentParser(description="Update license headers")
    parser.add_argument("--header", default="license.header", help="header file")
    parser.add_argument(
        "--extra",
        default=80,
        help="extra characters past beginning of file to look for header",
    )
    parser.add_argument(
        "--editdist", default=12, type=int, help="max edit distance between headers"
    )
    parser.add_argument(
        "--remove", default=False, action="store_true", help="remove the header"
    )
    parser.add_argument(
        "--cslash",
        default=False,
        action="store_true",
        help='use C slash "//" style comments',
    )
    parser.add_argument(
        "-v", default=False, action="store_true", dest="verbose", help="verbose output"
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-k", default=False, action="store_true", dest="check", help="check headers"
    )
    group.add_argument(
        "-i",
        default=False,
        action="store_true",
        dest="inplace",
        help="edit file inplace",
    )

    parser.add_argument("files", metavar="FILES", nargs="+", help="files to process")

    return parser.parse_args()


def file_read(filename):
    with open(filename) as file:
        return file.read()


def file_lines(filename):
    return file_read(filename).rstrip().split("\n")


def wrapper(prefix, leader, suffix, header):
    return prefix + "\n".join([leader + line for line in header]) + suffix


def wrapper_chpp(header, args):
    if args.cslash:
        return wrapper("", "//", "\n", header)
    else:
        return wrapper("/*\n", " *", "\n */\n", header)


def wrapper_hash(header, args):
    return wrapper("", "#", "\n", header)


file_types = OrderedDict(
    {
        "CMakeLists.txt": attrdict({"wrapper": wrapper_hash, "hashbang": False}),
        "Makefile": attrdict({"wrapper": wrapper_hash, "hashbang": False}),
        "*.cmake": attrdict({"wrapper": wrapper_hash, "hashbang": False}),
        "*.cpp": attrdict({"wrapper": wrapper_chpp, "hashbang": False}),
        "*.hpp": attrdict({"wrapper": wrapper_chpp, "hashbang": False}),
        "*.dockfile": attrdict({"wrapper": wrapper_hash, "hashbang": False}),
        "*.h": attrdict({"wrapper": wrapper_chpp, "hashbang": False}),
        "*.inc": attrdict({"wrapper": wrapper_chpp, "hashbang": False}),
        "*.java": attrdict({"wrapper": wrapper_chpp, "hashbang": False}),
        "*.prolog": attrdict({"wrapper": wrapper_chpp, "hashbang": False}),
        "*.py": attrdict({"wrapper": wrapper_hash, "hashbang": True}),
        "*.sh": attrdict({"wrapper": wrapper_hash, "hashbang": True}),
        "*.thrift": attrdict({"wrapper": wrapper_chpp, "hashbang": False}),
        "*.txt": attrdict({"wrapper": wrapper_hash, "hashbang": True}),
        "*.yml": attrdict({"wrapper": wrapper_hash, "hashbang": False}),
        "*.cu": attrdict({"wrapper": wrapper_chpp, "hashbang": False}),
        "*.cuh": attrdict({"wrapper": wrapper_chpp, "hashbang": False}),
        "*.clcpp": attrdict({"wrapper": wrapper_chpp, "hashbang": False}),
        "*.mm": attrdict({"wrapper": wrapper_chpp, "hashbang": False}),
        "*.metal": attrdict({"wrapper": wrapper_chpp, "hashbang": False}),
    }
)

file_pattern = regex.compile(
    "|".join(["^" + fnmatch.translate(type) + "$" for type in file_types.keys()])
)


def get_filename(filename):
    return os.path.basename(filename)


def get_fileextn(filename):
    split = os.path.splitext(filename)
    if len(split) <= 1:
        return ""

    return split[-1]


def get_wrapper(filename):
    if filename in file_types:
        return file_types[filename]

    return file_types["*" + get_fileextn(filename)]


def message(file, string):
    if file:
        print(string, file=file)


def main():
    fail = False
    log_to = None

    args = parse_args()

    if args.verbose:
        log_to = sys.stderr

    if args.check:
        log_to = None

        if args.verbose:
            log_to = sys.stdout

    header_text = file_lines(args.header)

    if len(args.files) == 1 and args.files[0] == "-":
        files = [file.strip() for file in sys.stdin.readlines()]
    else:
        files = args.files

    for filepath in files:
        filename = get_filename(filepath)

        matched = file_pattern.match(filename)

        if not matched:
            message(log_to, "Skip : " + filepath)
            continue

        content = file_read(filepath)
        wrap = get_wrapper(filename)

        header_comment = wrap.wrapper(header_text, args)

        start = 0
        end = 0

        # Look for an exact substr match
        #
        found = content.find(header_comment, 0, len(header_comment) + args.extra)
        if found >= 0:
            if not args.remove:
                message(log_to, "OK   : " + filepath)
                continue

            start = found
            end = found + len(header_comment)
        else:
            # Look for a fuzzy match in the first 60 chars
            #
            found = regex.search(
                "(?be)(%s){e<=%d}" % (regex.escape(header_comment[0:60]), 6),
                content[0 : 80 + args.extra],
            )
            if found:
                fuzzy = regex.compile(
                    "(?be)(%s){e<=%d}" % (regex.escape(header_comment), args.editdist)
                )

                # If the first 80 chars match - try harder for the rest of the header
                #
                found = fuzzy.search(
                    content[0 : len(header_comment) + args.extra], found.start()
                )
                if found:
                    start = found.start()
                    end = found.end()

        if args.remove:
            if start == 0 and end == 0:
                if not args.inplace:
                    print(content, end="")

                message(log_to, "OK   : " + filepath)
                continue

            # If removing the header text, zero it out there.
            header_comment = ""

        message(log_to, "Fix  : " + filepath)

        if args.check:
            fail = True
            continue

        # Remove any partially matching header
        #
        content = content[0:start] + content[end:]

        if wrap.hashbang:
            search = regex.search("^#!.*\n", content)
            if search:
                content = (
                    content[search.start() : search.end()]
                    + header_comment
                    + content[search.end() :]
                )
            else:
                content = header_comment + content
        else:
            content = header_comment + content

        if args.inplace:
            with open(filepath, "w") as file:
                print(content, file=file, end="")
        else:
            print(content, end="")

    if fail:
        return 1

    return 0


if __name__ == "__main__":
    sys.exit(main())
